
from sklearn.preprocessing.data import QuantileTransformer as SKLModel
import lale.helpers
import lale.operators
from numpy import nan, inf

class QuantileTransformerImpl():

    def __init__(self, n_quantiles=1000, output_distribution='uniform', ignore_implicit_zeros=False, subsample=100000, random_state=None, copy=True):
        self._hyperparams = {
            'n_quantiles': n_quantiles,
            'output_distribution': output_distribution,
            'ignore_implicit_zeros': ignore_implicit_zeros,
            'subsample': subsample,
            'random_state': random_state,
            'copy': copy}

    def fit(self, X, y=None):
        self._sklearn_model = SKLModel(**self._hyperparams)
        if (y is not None):
            self._sklearn_model.fit(X, y)
        else:
            self._sklearn_model.fit(X)
        return self

    def transform(self, X):
        return self._sklearn_model.transform(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for QuantileTransformer    Transform features using quantiles information.',
    'allOf': [{
        'type': 'object',
        'required': ['n_quantiles', 'output_distribution', 'ignore_implicit_zeros', 'subsample', 'random_state', 'copy'],
        'relevantToOptimizer': ['n_quantiles', 'output_distribution', 'subsample', 'copy'],
        'additionalProperties': False,
        'properties': {
            'n_quantiles': {
                'type': 'integer',
                'minimumForOptimizer': 1000,
                'maximumForOptimizer': 1001,
                'distribution': 'uniform',
                'default': 1000,
                'description': 'Number of quantiles to be computed. It corresponds to the number'},
            'output_distribution': {
                'enum': ['normal', 'uniform'],
                'default': 'uniform',
                'description': 'Marginal distribution for the transformed data. The choices are'},
            'ignore_implicit_zeros': {
                'type': 'boolean',
                'default': False,
                'description': 'Only applies to sparse matrices. If True, the sparse entries of the'},
            'subsample': {
                'type': 'integer',
                'minimumForOptimizer': 1,
                'maximumForOptimizer': 100000,
                'distribution': 'uniform',
                'default': 100000,
                'description': 'Maximum number of samples used to estimate the quantiles for'},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'If int, random_state is the seed used by the random number generator;'},
            'copy': {
                'type': 'boolean',
                'default': True,
                'description': 'Set to False to perform inplace transformation and avoid a copy (if the'},
        }}, {
        'XXX TODO XXX': 'Parameter: ignore_implicit_zeros > only applies to sparse matrices'}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Compute the quantiles used for transforming.',
    'type': 'object',
    'properties': {
        'X': {
            'XXX TODO XXX': 'ndarray or sparse matrix, shape (n_samples, n_features)',
            'description': 'The data used to scale along the features axis. If a sparse'},
    },
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Feature-wise transformation of the data.',
    'type': 'object',
    'properties': {
        'X': {
            'XXX TODO XXX': 'ndarray or sparse matrix, shape (n_samples, n_features)',
            'description': 'The data used to scale along the features axis. If a sparse'},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'The projected data.',
    'XXX TODO XXX': 'ndarray or sparse matrix, shape (n_samples, n_features)',
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema},
}
if (__name__ == '__main__'):
    lale.helpers.validate_is_schema(_combined_schemas)
QuantileTransformer = lale.operators.make_operator(QuantileTransformerImpl, _combined_schemas)




# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Set, Iterable, Iterator, Optional, Tuple, Union, Callable

# Type definitions
Schema = Any
SchemaEnum = Set[Any]

STrue:Schema = {}
SFalse:Schema = {"not":STrue}

def is_true_schema(s:Schema)->bool:
    return s is True or s == STrue

def is_false_schema(s:Schema)->bool:
    return s is False or s == SFalse

def getForOptimizer(obj, prop:str):
    return obj.get(prop + 'ForOptimizer', None)

def getMinimum(obj):
    prop = 'minimum'
    m = obj.get(prop)
    mfo = getForOptimizer(obj, prop)
    if mfo is None:
        return m
    else:
        if m is not None and mfo < m:
            raise ValueError(f"A minimum ({m}) and a *smaller* minimumForOptimizer ({mfo}) was specified in {obj}")
        return mfo

def getMaximum(obj):
    prop = 'maximum'
    m = obj.get(prop)
    mfo = getForOptimizer(obj, prop)
    if mfo is None:
        return m
    else:
        if m is not None and mfo > m:
            raise ValueError(f"A maximum ({m}) and a *greater* maximumForOptimizer ({mfo}) was specified in {obj}")
        return mfo

def getExclusiveMinimum(obj):
    prop = 'exclusveMinimum'
    m = obj.get(prop)
    mfo = getForOptimizer(obj, prop)
    if mfo is None:
        return m
    else:
        return mfo

def getExclusiveMaximum(obj):
    prop = 'exclusiveMaximum'
    m = obj.get(prop)
    mfo = getForOptimizer(obj, prop)
    if mfo is None:
        return m
    else:
        return mfo

def forOptimizer(s:Schema)->bool:
    if isinstance(s, dict):
        return s.get('forOptimizer', True)
    else:
        return True



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lale.helpers
import lale.operators
import pandas as pd

class NoOpImpl():
    def __init__(self, hyperparams=None):
        self._hyperparams = hyperparams

    # def fit(self, X, y = None):
    #     result = NoOpImpl(self._hyperparams)
    #     if isinstance(X, pd.DataFrame):
    #         result._feature_names = list(X.columns)
    #     else:
    #         #This assumes X is a 2d array which is consistent with its schema.
    #         result._feature_names = ['x%d' % i for i in range(X.shape[1])] 
    #     return result

    def transform(self, X, y = None):
        return X, y

    def transform_schema(self, s_X):
        return s_X

    # def get_feature_names(self, input_features=None):
    #     if input_features is not None:
    #         return list(input_features)
    #     elif self._feature_names is not None:
    #         return self._feature_names
    #     else:
    #         raise ValueError('Can only call get_feature_names on a trained operator. Please call fit to get a trained operator.')

_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Hyperparameter schema for the NoOp, which is a place-holder for no operation.',
    'allOf': [
    {   'description': 'This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters',
        'type': 'object',
        'additionalProperties': False,
        'relevantToOptimizer': [],
        'properties': {}}]}

_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Input data schema for training NoOp.',
    'type': 'object',
    'required': ['X'],
    'additionalProperties': False,
    'properties': {
        'X': {}}}

_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Input data schema for transformations using NoOp.',
    'type': 'object',
    'required': ['X', 'y'],
    'additionalProperties': False,
    'properties': {
        'X': {},
        'y': {}}}

_output_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Output data schema for transformations using NoOp.'}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.lale.no_op.html',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output': _output_schema }}

if (__name__ == '__main__'):
    lale.helpers.validate_is_schema(_combined_schemas)

NoOp = lale.operators.make_operator(NoOpImpl, _combined_schemas)



from torch.utils.data import Dataset

class NumpyTorchDataset(Dataset):
    """Pytorch Dataset subclass that takes a numpy array and an optional label array."""

    def __init__(self, X, y=None):
        """X and y are the dataset and labels respectively.
        
        Parameters
        ----------
        X : numpy array
            Two dimensional dataset of input features.
        y : numpy array
            Labels
        """
        self.X = X
        self.y = y

    def __len__(self):
        return self.X.shape[0]

    def __getitem__(self, idx):
        label = None
        if self.y is not None:
            return self.X[idx], self.y[idx]
        else:
            return self.X[idx]

    def get_data(self):
        if self.y is None:
            return self.X
        else:
            return self.X, self.y                


# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import logging
import random
import tensorflow_hub as hub
import tensorflow as tf
import numpy as np
from lale.operators import make_operator
import lale.helpers

logging.basicConfig(level=logging.INFO)


class USEPretrainedEncoderImpl(object):
    """
    USEPretrainedEncoderImpl is a module that allows simple consumption and fine-tuning of the
    DAN version of Universal Sentence Encoder: https://tfhub.dev/google/universal-sentence-encoder/2

    Parameters
    ----------
    model_path: string, (default=None), path to save the model

    batch_size: int, (default=32), batch size of fine-tuning the model

    References
    ----------
    Daniel Cer, et al. Universal Sentence Encoder. arXiv:1803.11175, 2018
    """
    def __init__(self,
                 model_path=None,
                 batch_size=32):
        self.resources_dir = os.path.join(os.path.dirname(__file__), 'resources')

        if model_path is None:
            model_path = os.path.join('pretrained_USE', '1fb57c3ffe1a38479233ee9853ddd7a8ac8a8c47')
        if os.path.exists(os.path.join(self.resources_dir, model_path)):
            self.url = os.path.join(self.resources_dir, model_path)
        else:
            os.environ['TFHUB_CACHE_DIR'] = os.path.join(self.resources_dir, 'pretrained_USE')
            self.url = "https://tfhub.dev/google/universal-sentence-encoder/2"

        # load the use model from saved location or tensorflow hub
        self.embed = hub.Module(self.url, trainable=True)
        self.sess = tf.Session()
        self.epochs = 10
        self.lr = 0.01
        self.batch_size = batch_size
        self.sess.run([tf.global_variables_initializer(),
                       tf.tables_initializer()])

    def fit(self, X, y, model_dir=None):
        """
        method for fine-tuning the universal sentence encoder using text classification task;
        fine-tune the current model and save the fine_tuned model for later application

        Parameters
        ----------
        X : list of str, input corpus for fine-tune use

        y : list of int, input label for fine-tune use

        model_dir: str, (default = None), directory to save the fine_tuned model for later use

        Returns
        -------

        """
        Y = np.array(y).reshape(-1)
        num_classes = len(np.unique(Y))
        Y = np.eye(num_classes)[Y]

        def next_batch(x, y, current_step, batch_size=self.batch_size):
            """method for extracton of next batch of data"""
            if len(x) < batch_size:
                data_idx = np.arange((len(x))).tolist()
                random.shuffle(data_idx)
                x_batch = [x[idx] for idx in data_idx]
                y_batch = [y[idx] for idx in data_idx]
            else:
                if current_step+batch_size < len(x):
                    x_batch = x[current_step:current_step+batch_size]
                    y_batch = y[current_step:current_step+batch_size]
                    current_step += batch_size
                else:
                    x_batch = x[current_step:]
                    y_batch = y[current_step:]

                    data_idx = np.arange((len(x))).tolist()
                    random.shuffle(data_idx)
                    x = [x[idx] for idx in data_idx]
                    y = [y[idx] for idx in data_idx]
                    current_step = 0

            return x, y, x_batch, y_batch, current_step

        X_batch = tf.placeholder(dtype=tf.string, shape=(None))
        Y_batch = tf.placeholder(dtype=tf.int16, shape=(None, num_classes))
        embedded_representation = self.embed(X_batch)
        dense_layer = tf.layers.Dense(units=num_classes, use_bias=True)
        logits = dense_layer(embedded_representation)
        los = tf.nn.softmax_cross_entropy_with_logits_v2(labels=Y_batch, logits=logits)
        optimizer = tf.train.AdamOptimizer(learning_rate=self.lr)
        train_opt = optimizer.minimize(loss=los)

        curr = 0
        iterations = int(np.ceil(len(X)/self.batch_size))
        self.sess.run(tf.global_variables_initializer())
        for _ in range(self.epochs):
            for _ in range(iterations):
                X, Y, x_batch, y_batch, curr = next_batch(X, Y, curr)
                self.sess.run(train_opt, feed_dict={X_batch: x_batch, Y_batch: y_batch})

        # save model
        if model_dir is None:
            model_dir = os.path.join(self.resources_dir, 'fine_tuned_USE')

        if os.path.exists(model_dir):
            shutil.rmtree(model_dir)
        self.embed.export(model_dir, session=self.sess)

        return self

    def transform(self, X):
        """
        method for encoding strings into floating point arrays using universal sentence encoder
        Parameters
        ----------
        X : list of strings, input corpus for fine-tune use

        Returns
        -------
        transformed_x: 2d-array, shape [n_samples, 512], sentence embedding for downstream task
        """
        sentence_embedding = self.embed(X)
        transformed_x = self.sess.run(sentence_embedding)
        return transformed_x


_input_schema_fit = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Input data schema for training.',
    'type': 'object',
    'required': ['X', 'y'],
    'additionalProperties': False,
    'properties': {
        'X': {
            'description': 'Input Text',
            'anyOf': [
                {'type': 'array',
                 'items': {'type': 'string'}},
                {'type': 'array',
                 'items': {
                     'type': 'array', 'minItems': 1, 'maxItems': 1,
                     'items': {'type': 'string'}}}]},
        'y': {
            'description': 'Labels, required',
            'anyOf': [
                {'type': 'array',
                 'items': {'type': 'number'}}
            ]
        }
    }}

_input_schema_predict = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Input data schema for training.',
    'type': 'object',
    'required': ['X'],
    'additionalProperties': False,
    'properties': {
        'X': {
            'description': 'Input Text',
            'anyOf': [
                {'type': 'array',
                 'items': {'type': 'string'}},
                {'type': 'array',
                 'items': {
                     'type': 'array', 'minItems': 1, 'maxItems': 1,
                     'items': {'type': 'string'}}}]}
    }}

_output_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Output data schema for transformed data.',
    'type': 'array',
    'items': {'type': 'array', 'items': {'type': 'number'}}}

_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Hyperparameter schema.',
    'allOf': [
        {'description':
             'This first sub-object lists all constructor arguments with their '
             'types, one at a time, omitting cross-argument constraints.',
         'type': 'object',
         'additionalProperties': False,
         'relevantToOptimizer': [],
         'properties': {
             'batch_size': {
                 'description': 'Batch size used for transform.',
                 'type': 'integer',
                 'default': 32,
                 'minimum': 1,
                 'distribution': 'uniform',
                 'minimumForOptimizer': 16,
                 'maximumForOptimizer': 128}}}]}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters for a transformer for'
                   ' a text data transformer based on pre-trained USE model '
                   '(https://tfhub.dev/google/universal-sentence-encoder/2).',
    'type': 'object',
    'tags': {
        'pre': ['text'],
        'op': ['transformer', '~interpretable'],
        'post': ['embedding']},
    'properties': {
        'input_fit': _input_schema_fit,
        'input_predict': _input_schema_predict,
        'output': _output_schema,
        'hyperparams': _hyperparams_schema}}

if __name__ == "__main__":
    lale.helpers.validate_is_schema(_combined_schemas)

USEPretrainedEncoder = make_operator(USEPretrainedEncoderImpl, _combined_schemas)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .concat_features import ConcatFeatures
from .no_op import NoOp
from .hyperopt_classifier import HyperoptClassifier
from .sample_based_voting import SampleBasedVoting
from .hyperopt_regressor import HyperoptRegressor
from .keep_non_numbers import KeepNonNumbers
from .keep_numbers import KeepNumbers
from .project import Project
from .batching import Batching



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import itertools
import jsonschema

from .schema_ranges import SchemaRange

from typing import Any, Dict, Generic, List, Set, Iterable, Iterator, Optional, Tuple, TypeVar, Union
from .schema_utils import Schema, getMinimum, getMaximum, forOptimizer, STrue, SFalse, is_true_schema, is_false_schema

logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)

# Goal: given a json schema, convert it into an equivalent json-schema
# in "grouped-dnf" form:
# allOf: [anyOf: nochoice], where
# nochoice
# 
# initial version, which does not try to group things intelligently:
# allOf [anyOf [P1 P2], anyOf[Q1 Q2]] ==
# anyOf [map allOf [Ps]x[Pqs]]
# Note that P1 == anyOf [P] == allOf [P]

# Given a schema, if it is an anyof, return the list of choices.
# Otherwise, return a singleton choice -- the schema

# enumerations should logically be sets.
# However, the keys are not hashable
VV = TypeVar("VV")
class set_with_str_for_keys(Generic[VV]):
    """ This mimicks a set, but uses the string representation
        of the elements for comparison tests.
        It can be used for unhashable elements, as long 
        as the str function is injective
    """
    _elems:Dict[str,VV]
    def __init__(self, elems:Union[Dict[str,VV], Iterator[VV]]):
        if isinstance(elems, dict):
            self._elems = elems
        else:
            self._elems = {str(v):v for v in elems}

    def __iter__(self):
        return iter(self._elems.values())

    def __bool__(self):
        return bool(self._elems)

    def __str__(self):
        return str(list(self._elems.values()))

    def union(self, *others):
        return set_with_str_for_keys([elem for subl in [self]+list(others) for elem in subl])

    def intersection(self, *others):
        d:Dict[str,VV] = dict(self._elems)
        for ssk in others:
            for k in list(d.keys()):
                if k not in ssk._elems:
                    del d[k]
        return set_with_str_for_keys(d)

    def difference(self, *others):
        d:Dict[str,VV] = dict(self._elems)
        for ssk in others:
            for k in list(d.keys()):
                if k in ssk._elems:
                    del d[k]
        return set_with_str_for_keys(d)
    

def toAnyOfList(schema:Schema)->List[Schema]:
    if 'anyOf' in schema:
        return schema['anyOf']
    else:
        return [schema]

def toAllOfList(schema:Schema)->List[Schema]:
    if 'allOf' in schema:
        return schema['allOf']
    else:
        return [schema]

def liftAllOf(schemas:List[Schema])->Iterator[Schema]:
    """ Given a list of schemas, if any of them are 
        allOf schemas, lift them out to the top level
    """
    for sch in schemas:
        schs2 = toAllOfList(sch)
        for s in schs2:
            yield s


def liftAnyOf(schemas:List[Schema])->Iterator[Schema]:
    """ Given a list of schemas, if any of them are 
        anyOf schemas, lift them out to the top level
    """
    for sch in schemas:
        schs2 = toAnyOfList(sch)
        for s in schs2:
            yield s

def makeSingleton_(k:str, schemas:List[Schema])->Schema:
    if len(schemas) == 0:
        return {}
    if len(schemas) == 1:
        return schemas[0]
    else:
        return {k:schemas}

def makeAllOf(schemas:List[Schema])->Schema:
    return makeSingleton_('allOf', schemas)
def makeAnyOf(schemas:List[Schema])->Schema:
    return makeSingleton_('anyOf', schemas)
def makeOneOf(schemas:List[Schema])->Schema:
    return makeSingleton_('oneOf', schemas)

# This is a great function for a breakpoint :-)
def impossible()->Schema:
    return SFalse

def enumValues(es:set_with_str_for_keys[Any], s:Schema)->set_with_str_for_keys[Any]:
    """Given an enumeration set and a schema, return all the consistent values of the enumeration."""
    # TODO: actually check.  This should call the json schema validator
    ret = list()
    try:
        for e in es:
            try:
                jsonschema.validate(e, s)
                ret.append(e)
            except:
                logger.debug(f"enumValues: {e} removed from {es} because it does not validate according to {s}")
        return set_with_str_for_keys(iter(ret))
    except jsonschema.ValidationError as error:
        logger.warning(f"enumValues: Schema {s} does not validate: {error}")
        # Let us be conservative
        return es


# invariants for all the simplify* functions:
## invariant: if floatAny then at most the top level return value will be 'anyOf'
## invariant: if there is no (nested or top level) 'anyOf' then the result will not have any either

extra_field_names:List[str] = ['default', 'description']


def simplifyAll(schemas:List[Schema], floatAny:bool)->Schema:
    # First, we partition the schemas into the different types
    # that we care about
    combined_original_schema:Schema = {'allOf':schemas}
    s_all:List[Schema] = schemas
    s_any:List[List[Schema]] = []
    s_one:List[Schema] = []

    s_not:List[Schema] = []
    s_not_number_list:List[Schema] = [] # a list of schemas that are a top level 'not' with a type='integer' or 'number' under it

    s_not_enum_list:List[set_with_str_for_keys[Any]] = []
    s_enum_list:List[set_with_str_for_keys[Any]] = []


    s_type = None
    s_typed = []
    s_other = []
    s_not_for_optimizer:List[Schema] = []

    while s_all:
        l = s_all
        s_all = []
        for s in l:
            if s is None:
                continue
            s = simplify(s, floatAny)
            if s is None:
                continue
            if not forOptimizer(s):
                logger.info(f"simplifyAll: skipping not for optimizer {s} (after simplification)")
                s_not_for_optimizer.append(s)
                continue
            if is_true_schema(s):
                continue
            if is_false_schema(s):
                return SFalse
            if 'allOf' in s:
                s_all.extend(s['allOf'])
            elif 'anyOf' in s:
                s_any.append(s['anyOf'])
            elif 'oneOf' in s:
                s_one.append(s)
            elif 'not' in s:
                snot = s['not']
                if snot is None:
                    continue
                elif 'enum' in snot:
                    ev = enumValues(set_with_str_for_keys(snot['enum']), {'not':combined_original_schema})
                    s_not_enum_list.append(ev)
                elif 'type' in snot and (snot['type'] == 'number' or snot['type'] == 'integer'):
                    s_not_number_list.append(s)
                else:
                    s_not.append(s)
            elif 'enum' in s:
                # TODO: copy over extra fields (description...)
                ev = enumValues(set_with_str_for_keys(s['enum']), combined_original_schema)
                if ev:
                    s_enum_list.append(ev)
                else:
                    logger.info(f"simplifyAll: {schemas} is not a satisfiable list of conjoined schemas because the enumeration {list(s['enum'])} has no elements that are satisfiable by the conjoined schemas")
                    return impossible()
            elif 'type' in s:
                t = s['type']
                if s_type:
                    if s_type == 'number' and t == 'integer' or s_type == 'integer' and t == 'number':
                        s_type = 'integer'
                    elif s_type != t:
                        logger.info(f"simplifyAll: {schemas} is not a satisfiable list of conjoined schemas because {s} has type '{t}' and a previous schema had type '{s_type}'")
                        return impossible()
                else:
                    s_type = t
                s_typed.append(s)
            elif 'XXX TODO XXX' in s and len(s) == 1:
                # Ignore missing constraints
                pass
            else:
                logger.warning(f"simplifyAll: '{s}' has unknown type")
                s_other.append(s)
    # Now that we have partitioned things 
    # Note: I am sure some of our assumptions here are not correct :-(, but this should do for now :-)

    # let's try to find a quick contradiction
    if s_not or s_not_number_list:
        # a bit of a special case here (which should eventually be replaced by more prinicipalled logic):
        # if one of the not cases is identical to to one of the extra cases
        # then this entire case is impossible.
        # This provides a workaround to #42 amongst other problems

        # first gather the set of extras
        pos_k = set()
        for k in s_typed:
            pos_k.add(str(k))

        for sn in itertools.chain(s_not, s_not_number_list):
            snn = sn['not']
            if str(snn) in pos_k:
                logger.info(f"simplifyAll: Contradictory schema {str(combined_original_schema)} contains both {str(snn)} and its negation")
                return impossible()


    # first, we simplify enumerations

    s_enum:Optional[set_with_str_for_keys[Any]] = None
    s_not_enum:Optional[set_with_str_for_keys[Any]] = None

    if s_enum_list:
        # if there are enumeration constraints, we want their intersection
        s_enum = set_with_str_for_keys.intersection(*s_enum_list)
        if not s_enum:
            # This means that enumeration values where specified
            # but none are possible, so this schema is impossible to satisfy
            logger.info(f"simplifyAll: {schemas} is not a satisfiable list of conjoined schemas because the conjugation of these enumerations {list(s_enum_list)} is unsatisfiable (the intersection is empty)")
            return impossible()
    if s_not_enum_list:
        s_not_enum = set_with_str_for_keys.union(*s_not_enum_list)
    
    if s_enum and s_not_enum:
        s_enum_diff = set_with_str_for_keys.difference(s_enum, s_not_enum)
        if not s_enum_diff:
            # This means that enumeration values where specified
            # but none are possible, so this schema is impossible to satisfy
            logger.info(f"simplifyAll: {schemas} is not a satisfiable list of conjoined schemas because the conjugation of the enumerations is {s_enum} all of which are excluded by the conjugation of the disallowed enumerations {s_not_enum}")
            return impossible()
        s_enum = s_enum_diff
        s_not_enum = None

    s_extra:Dict[str,Any] = {}

    # break out, combine, and keep 'extra' fields, like description
    if s_typed:
        s_typed = [s.copy() for s in s_typed]
        for o in s_typed:
            for k in extra_field_names:
                if k in o:
                    d = o[k]
                    if k in s_extra and s_extra[k] != d:
                        logger.info(f"mergeAll: conflicting {k} fields: {s_extra[k]} and {d} found when merging schemas {schemas}")
                    else:
                        s_extra[k] = d
                    del o[k]
        s_typed = [s for s in s_typed if s]

    if s_type == 'number' or s_type == 'integer':
        # First we combine all the positive number range schemas
        s_range = SchemaRange()
        s_range_for_optimizer = SchemaRange()

        for o in s_typed:
            o_range = SchemaRange.fromSchema(o)
            s_range &= o_range

            o_range_for_optimizer = SchemaRange.fromSchemaForOptimizer(o)
            s_range_for_optimizer &= o_range_for_optimizer
        # now let us look at negative number ranges
        # for now, we will not handle cases that would require splitting ranges
        # TODO: 42 is about handling more reasoning
        s_not_list = s_not_number_list
        s_not_number_list = []
        for s in s_not_list:
            snot = s['not']
            o_range = SchemaRange.fromSchema(snot)

            success = s_range.diff(o_range)
            if success is None:
                logger.info(f"simplifyAll: [range]: {s} is not a satisfiable schema, since it negates everything, falsifying the entire combined schema {combined_original_schema}")
                return impossible()
            o_range_for_optimizer = SchemaRange.fromSchemaForOptimizer(snot)
            success2 = s_range_for_optimizer.diff(o_range_for_optimizer)
            if success2 is None:
                logger.info(f"simplifyAll: [range]: {s} is not a satisfiable schema for the optimizer, since it negates everything, falsifying the entire combined schema {combined_original_schema}")
                return impossible()

            elif success is False or success2 is False:
                s_not_number_list.append(s)

        # Now we look at negative enumarations.
        # for now, we will not handle cases that would require splitting ranges
        # TODO: 42 is about handling more reasoning
        if s_not_enum:
            s_cur_not_enum_list:set_with_str_for_keys[Any] = s_not_enum
            s_not_enum_l:List[Any] = []

            for s in s_cur_not_enum_list:
                if isinstance(s, (int,float)):
                    success = s_range.remove_point(s)
                    if success is None:
                        logger.info(f'simplifyAll: [range]: {{"not": {{"enum": [{s}]}}}}  is not a satisfiable schema, since it negates everything, falsifying the entire combined schema {combined_original_schema}')
                        return impossible()
                    success2 = s_range_for_optimizer.remove_point(s)
                    if success2 is None:
                        logger.info(f'simplifyAll: [range]: {{"not": {{"enum": [{s}]}}}}  is not a satisfiable schema for the optimizer, since it negates everything, falsifying the entire combined schema {combined_original_schema}')
                        return impossible()
                    
                    elif success is False or success2 is False:
                        s_not_enum_l.append(s)
            s_not_enum = set_with_str_for_keys(iter(s_not_enum_l))



        # now let us put everything back together
        number_schema = SchemaRange.to_schema_with_optimizer(s_range, s_range_for_optimizer)

        if SchemaRange.is_empty2(s_range, s_range):
            logger.info(f"simplifyAll: [range]: range simplification determined that the required minimum is greater than the required maximum, so the entire thing is unsatisfiable {combined_original_schema}")
            # if the actual range is empty, the entire schema is invalid
            return impossible()
        elif SchemaRange.is_empty2(s_range_for_optimizer, s_range):
            number_schema['forOptimizer'] = SFalse
            logger.info(f"simplifyAll: [range]: range simplification determined that the required minimum for the optimizer is greater than the required maximum, so the range is being marked as not for the optimizer: {number_schema}")
        elif SchemaRange.is_empty2(s_range, s_range_for_optimizer):
            number_schema['forOptimizer'] = SFalse
            logger.info(f"simplifyAll: [range]: range simplification determined that the required minimum is greater than the required maximum for the optimizer, so the range is being marked as not for the optimizer: {number_schema}")
        elif SchemaRange.is_empty2(s_range_for_optimizer, s_range_for_optimizer):
            logger.info(f"simplifyAll: [range]: range simplification determined that the required minimum for the optimizer is greater than the required maximum for the optimizer, so the range is being marked as not for the optimizer: {number_schema}")
            number_schema['forOptimizer'] = SFalse

        s_typed = [number_schema]

    elif s_type == 'object':
        # if this is an object type, we want to merge the properties
        s_required:Set[str] = set()
        s_props:Dict[str, Schema] = {}
        # TODO: generalize this to handle schema types here
        s_additionalProperties = True
        # propertyNames = []

        for o in s_typed:
            o_required = o.get('required', None)
            if o_required:
                s_required = s_required.union(o_required)

            # TODO: handle empty/absent properties case
            if 'properties' in o:
                o_props = o['properties']
            else:
                o_props = {}
            o_additionalProperties = 'additionalProperties' not in o or o['additionalProperties']
            # safety check:
            if not o_additionalProperties:
                for p in s_required:
                    if p not in o_props:
                        # There is a required key, but our schema
                        # does not contain that key and does not allow additional properties
                        # This schema can never be satisfied, so we can simplify this whole thing to the False schema
                        logger.info(f"simplifyAll: {s_typed} is not a mergable list of schemas because {o} does not have the required key '{p}' and excludes additional properties")
                        return impossible()

            # If we do not allow additional properties
            # Remove all existing properties that are 
            # not in our schema
            if not o_additionalProperties:
                for p in s_props:
                    if p not in o_props:
                        del s_props[p]
            # now go through our properties and add them
            for p,pv in o_props.items():
                if p in s_props:
                    s_props[p].append(pv)
                elif s_additionalProperties:
                    s_props[p] = [pv]
            s_additionalProperties = s_additionalProperties and o_additionalProperties
        # at this point, we have aggregated the object schemas
        # for all the properties in them
        if s_required and not s_additionalProperties:
            for k in s_required:
                if k not in s_props:
                    logger.info(f"simplifyAll: {s_typed} is not a mergable list of schemas because {o} requires key '{k}', which is not in earlier schemas, and an earlier schema excluded additional properties")
                    return impossible()

        merged_props = {p:simplifyAll(s_props[p], False) for p in s_props}
        if s_required:
            for k in s_required:
                if is_false_schema(merged_props.get(k, False)):
                    logger.info(f"simplifyAll: required key {k} is False, so the entire conjugation of schemas {schemas} is False")
                    return impossible()

        obj:Dict[Any,Any] = {}
        obj['type'] = 'object'
        if merged_props:
            obj['properties'] = merged_props
        if not s_additionalProperties:
            obj['additionalProperties'] = False
        if len(s_required) != 0:
            obj['required'] = list(s_required)
        s_typed = [obj]

    # TODO: more!
    assert not s_all
    ret_all = []
    ret_main = s_extra if s_extra else {}

    if s_enum:
        # we should simplify these as for s_not_enum
        ret_main['enum']=list(s_enum)
        return ret_main

    if ret_main:
        if s_typed:
            s_typed[0] = {**ret_main, **s_typed[0]}
        elif s_other:
            s_other[0] = {**ret_main, **s_other[0]}
        else:
            ret_all.append(ret_main)
    if s_typed:
        ret_all.extend(s_typed)
    if s_other:
        ret_all.extend(s_other)
    if s_not_for_optimizer:
        ret_all.extend(s_not_for_optimizer)
    if s_one:
        ret_all.extend(s_one)
    if s_not_number_list:
        ret_all.extend(s_not_number_list)
    if s_not:
        ret_all.extend(s_not)
    
    if s_not_enum:
        # We can't do not alongside anything else
        # TODO: we should validate the list against the 
        # other parts of ret_all (this would need to move down): if any elements don't validate
        # then they already would be excluded
        # we can simplify +enum's the same way
        ret_all_agg = makeAllOf(ret_all)
        s_not_enum_simpl = enumValues(s_not_enum, ret_all_agg)
        if s_not_enum_simpl:
            sne = {'not':{'enum':list(s_not_enum)}}
            ret_all.append(sne)
        else:
            logger.debug(f"simplifyAll: {s_not_enum} was a negated enum that was simplified away because its elements anyway don't satisfy the additional constraints {ret_all_agg}")

        s_not_enum = s_not_enum_simpl

    if not floatAny:
        ret_all.extend([simplifyAny(s, False) for s in s_any])
    ret_all_schema = makeAllOf(ret_all)
    if floatAny and s_any:
        args = list(([ret_all_schema], *tuple(s_any)))
        cp = list(itertools.product(*args))
        alls = [simplifyAll(list(s), False) for s in cp]
        ret = simplifyAny(alls, False)
        return ret
    else:
        return ret_all_schema

def simplifyAny(schema:List[Schema], floatAny:bool)->Schema:
    s_any = schema

    s_enum_list:List[set_with_str_for_keys[Any]] = []
    s_not_enum_list:List[set_with_str_for_keys[Any]] = []

    s_other:List[Schema] = []
    s_not_for_optimizer:List[Schema] = []

    while s_any:
        l = s_any
        s_any = []
        for s in l:
            if s is None:
                continue
            s = simplify(s, floatAny)
            if s is None:
                continue
            if not forOptimizer(s):
                logger.info(f"simplifyAny: skipping not for optimizer {s} (after simplification)")
                s_not_for_optimizer.append(s)
                continue
            if is_true_schema(s):
                return STrue
            if is_false_schema(s):
                continue
            if 'anyOf' in s:
                s_any.extend(s['anyOf'])
            elif 'enum' in s:
                ev = enumValues(set_with_str_for_keys(s['enum']), s)
                if ev:
                    s_enum_list.append(ev)
            elif 'not' in s:
                snot = s['not']
                if 'enum' in s['not']:
                    ev = enumValues(set_with_str_for_keys(snot['enum']), snot)
                    if ev:
                        s_not_enum_list.append(ev)
            else:
                s_other.append(s)

    s_enum:Optional[set_with_str_for_keys[Any]] = None
    s_not_enum:Optional[set_with_str_for_keys[Any]] = None

    if s_enum_list:
        # if there are enumeration constraints, we want their intersection
        s_enum = set_with_str_for_keys.union(*s_enum_list)

    if s_not_enum_list:
        s_not_enum = set_with_str_for_keys.intersection(*s_not_enum_list)
    
    if s_enum and s_not_enum:
        s_not_enum = set_with_str_for_keys.difference(s_not_enum, s_enum)
        s_enum = None

    
    assert not s_any
    ret:List[Schema] = []

    if s_enum:
        ret.append({'enum':list(s_enum)})
    if s_not_enum:
        ret.append({'not':{'enum':list(s_not_enum)}})
    ret.extend(s_other)
    ret.extend(s_not_for_optimizer)
    return makeAnyOf(ret)

def simplifyNot(schema:Schema, floatAny:bool)->Schema:
    return simplifyNot_(schema, floatAny, alreadySimplified=False)

def simplifyNot_(schema:Schema, floatAny:bool, alreadySimplified:bool=False)->Schema:   
    """alreadySimplified=true implies that schema has already been simplified"""
    if 'not' in schema:
        # if there is a not/not, we can just skip it
        ret = simplify(schema['not'], floatAny)
        return ret
    elif 'anyOf' in schema:
        anys = schema['anyOf']
        alls = [{'not':s} for s in anys]
        ret = simplifyAll(alls, floatAny)
        return ret
    elif 'allOf' in schema:
        alls = schema['allOf']
        anys = [{'not':s} for s in alls]
        ret = simplifyAny(anys, floatAny)
        return ret
    elif not alreadySimplified:
        s = simplify(schema, floatAny)
        # it is possible that the result of calling simplify
        # resulted in something that we can push 'not' down into
        # so we call ourselves, being careful to avoid an infinite loop.
        return simplifyNot_(s, floatAny, alreadySimplified=True)
    else:
        return {'not':schema}

def simplify(schema:Schema, floatAny:bool)->Schema:
    """ Tries to simplify a schema into an equivalent but
        more compact/simpler one.  If floatAny if true, then
        the only anyOf in the return value will be at the top level.
        Using this option may cause a combinatorial blowup in the size 
        of the schema
        """
    if is_true_schema(schema):
        return STrue
    if is_false_schema(schema):
        return SFalse
    if 'enum' in schema:
        # TODO: simplify the schemas by removing anything that does not validate
        # against the rest of the schema
        return schema
    if 'allOf' in schema:
        ret = simplifyAll(schema['allOf'], floatAny)
        return ret
    elif 'anyOf' in schema:
        ret = simplifyAny(schema['anyOf'], floatAny)
        return ret
    elif 'not' in schema:
        return simplifyNot(schema['not'], floatAny)
    elif 'type' in schema and schema['type'] == 'object' and 'properties' in schema:
        schema2 = schema.copy()
        props = {}
        all_objs = [schema2]
        ## TODO: how does this interact with required?
        ## {k1:s_1, k2:anyOf:[s2s], k3:anyOf:[s3s]}
        ## If floatAny is true and any properties have an anyOf in them
        ## we need to float it out to the top.  We can then
        ## give it to simplifyAll, which does the cross product to lift 
        ## them out of the list
        for k,v in schema['properties'].items():
            s = simplify(v, floatAny)
            if is_false_schema(s) and 'required' in schema and s in schema['required']:
                logger.info(f"simplify: required key {k} is False, so the entire schema {schema} is False")
                return impossible()

            if (not is_true_schema(s)) and floatAny and 'anyOf' in s:
                all_objs.append({'anyOf':[{'type':'object', 'properties':{k:vv}} for vv in s['anyOf']]})
                # If we are disallowing additionalProperties, then we can't remove this property entirely

                if not schema.get('additionalProperties', True):
                    props[k] = STrue
            else:
                props[k] = s
        schema2['properties'] = props
        if len(all_objs) == 1:
            return schema2
        else:
            # The termination argument here is somewhat subtle
            s = simplifyAll(all_objs, floatAny)
            return s
    else:
        return schema

# TODO: semantically, allOf should force an intersection
# of relevantFields, yet union seems kinder to the user/more modular (at least if additionalProperties:True)
def findRelevantFields(schema:Schema) -> Optional[Set[str]]:
    """Either returns the relevant fields for the schema, or None if there was none specified"""
    if 'allOf' in schema:
        fields_list:List[Optional[Set[str]]] = [findRelevantFields(s) for s in schema['allOf']]
        real_fields_list:List[Set[str]] = [f for f in fields_list if f is not None]
        if real_fields_list:
            return set.union(*real_fields_list)
        else:
            return None
    else:
        if 'relevantToOptimizer' in schema:
            return set(schema['relevantToOptimizer'])
        else:
            return None

    # does not handle nested objects and nested relevant fields well
def narrowToGivenRelevantFields(schema:Schema, relevantFields:Set[str])->Schema:
    if schema is False:
        return False
    if 'anyOf' in schema:
        return {'anyOf':[narrowToGivenRelevantFields(a, relevantFields) for a in schema['anyOf']]}
    if 'allOf' in schema:
        return {'allOf':[narrowToGivenRelevantFields(a, relevantFields) for a in schema['allOf']]}
    if 'not' in schema:
        return {'not':narrowToGivenRelevantFields(schema['not'], relevantFields)}
    if 'type' in schema and schema['type'] == "object" and 'properties' in schema:
        props = schema['properties']
        new_props = {k:narrowToGivenRelevantFields(v, relevantFields) for (k,v) in props.items() if k in relevantFields}
        schema2 = schema.copy()
        schema2['properties'] = new_props
        if 'required' in schema:
            reqs = set(schema['required'])
            schema2['required'] = list(reqs.intersection(relevantFields))
        return schema2
    else:
        return schema

def narrowToRelevantFields(schema:Schema)->Schema:
    relevantFields:Optional[Set[str]] = findRelevantFields(schema)
    if relevantFields is not None:
        return narrowToGivenRelevantFields(schema, relevantFields)
    else:
        return schema

# Given a json schema, removes any elements marked as 'forOptimizer:false'
# also does some basic simplifications
def filterForOptimizer(schema:Schema)->Optional[Schema]:
    if schema is None or is_true_schema(schema) or is_false_schema(schema):
        return schema
    if not forOptimizer(schema):
        return None
    if 'anyOf' in schema:
        subs = schema['anyOf']
        sch = [filterForOptimizer(s) for s in subs]
        sch_nnil = [s for s in sch if s is not None]
        if sch_nnil:
            return makeAnyOf(sch_nnil)
        else:
            return None
    if 'allOf' in schema:
        subs = schema['allOf']
        sch = [filterForOptimizer(s) for s in subs]
        sch_nnil = [s for s in sch if s is not None]
        filtered_sch = sch_nnil
        if len(sch_nnil) != len(sch):
            # Questionable semantics here (aka HACK!!!!)
            # Since we removed something from the schema
            # we will also remove negated schemas
            filtered_sch = [s for s in sch_nnil if not isinstance(s, dict) or 'not' not in s]

        if filtered_sch:
            return makeAllOf(filtered_sch)
        else:
            return None
    if 'oneOf' in schema:
        subs = schema['oneOf']
        sch = [filterForOptimizer(s) for s in subs]
        sch_nnil = [s for s in sch if s is not None]
        if sch_nnil:
            return makeOneOf(sch_nnil)
        else:
            return None

    if 'not' in schema:
        s = filterForOptimizer(schema['not'])
        if s is None:
            return None
        else:
            return {'not':s}
    if 'type' in schema and schema['type'] == 'object' and 'properties' in schema:
        required = schema.get('required', None)

        props = {}
        for k,v in schema['properties'].items():
            s = filterForOptimizer(v)
            if s is None:
#                if required and k in required:
                    # if this field is required (and has now been filtered)
                    # filter the whole object schema
                    return None
            else:
                props[k] = s
        
        ret = schema.copy()
        ret['properties'] = props
        return ret

    return schema

def narrowSimplifyAndFilter(schema:Schema, floatAny:bool)->Optional[Schema]:
    n_schema = narrowToRelevantFields(schema)
    simplified_schema = simplify(n_schema, floatAny)
    filtered_schema = filterForOptimizer(simplified_schema)
    return simplified_schema




# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from lale.lib.lale import Batching, NoOp
from lale.lib.sklearn import MinMaxScaler
from lale.lib.sklearn import MLPClassifier, LogisticRegression
from sklearn.metrics import accuracy_score

class TestBatching(unittest.TestCase):
    def setUp(self):
        from sklearn.datasets import load_iris
        from sklearn.model_selection import train_test_split
        data = load_iris()
        X, y = data.data, data.target
        self.X_train, self.X_test, self.y_train, self.y_test =  train_test_split(X, y)
           

    def test_fit(self):
        import warnings
        warnings.filterwarnings(action="ignore")
        from lale.lib.sklearn import MinMaxScaler, MLPClassifier
        pipeline = NoOp() >> Batching(operator = MinMaxScaler() >> MLPClassifier(random_state=42), batch_size = 112)
        trained = pipeline.fit(self.X_train, self.y_train)
        predictions = trained.predict(self.X_test)
        lale_accuracy = accuracy_score(self.y_test, predictions)

        from sklearn.preprocessing import MinMaxScaler
        from sklearn.neural_network import MLPClassifier
        prep = MinMaxScaler()
        trained_prep = prep.partial_fit(self.X_train, self.y_train)
        X_transformed = trained_prep.transform(self.X_train)

        clf = MLPClassifier(random_state=42)
        import numpy as np
        trained_clf = clf.partial_fit(X_transformed, self.y_train, classes = np.unique(self.y_train))
        predictions = trained_clf.predict(trained_prep.transform(self.X_test))
        sklearn_accuracy = accuracy_score(self.y_test, predictions)

        self.assertEqual(lale_accuracy, sklearn_accuracy)

    def test_fit1(self):
        import warnings
        warnings.filterwarnings(action="ignore")
        from lale.lib.sklearn import MinMaxScaler, MLPClassifier
        pipeline = Batching(operator = MinMaxScaler() >> MLPClassifier(random_state=42), batch_size = 112)
        trained = pipeline.fit(self.X_train, self.y_train)
        predictions = trained.predict(self.X_test)
        lale_accuracy = accuracy_score(self.y_test, predictions)

        from sklearn.preprocessing import MinMaxScaler
        from sklearn.neural_network import MLPClassifier
        prep = MinMaxScaler()
        trained_prep = prep.partial_fit(self.X_train, self.y_train)
        X_transformed = trained_prep.transform(self.X_train)

        clf = MLPClassifier(random_state=42)
        import numpy as np
        trained_clf = clf.partial_fit(X_transformed, self.y_train, classes = np.unique(self.y_train))
        predictions = trained_clf.predict(trained_prep.transform(self.X_test))
        sklearn_accuracy = accuracy_score(self.y_test, predictions)

        self.assertEqual(lale_accuracy, sklearn_accuracy)        

    def test_fit2(self):
        import warnings
        warnings.filterwarnings(action="ignore")
        from lale.lib.sklearn import MinMaxScaler, MLPClassifier
        pipeline = Batching(operator = MinMaxScaler() >> MinMaxScaler(), batch_size = 112)
        trained = pipeline.fit(self.X_train, self.y_train)
        lale_transforms = trained.transform(self.X_test)

        from sklearn.preprocessing import MinMaxScaler
        prep = MinMaxScaler()
        trained_prep = prep.partial_fit(self.X_train, self.y_train)
        X_transformed = trained_prep.transform(self.X_train)

        clf = MinMaxScaler()
        import numpy as np
        trained_clf = clf.partial_fit(X_transformed, self.y_train)
        sklearn_transforms = trained_clf.transform(trained_prep.transform(self.X_test))

        for i in range(5):
            for j in range(2):
                self.assertAlmostEqual(lale_transforms[i, j], sklearn_transforms[i, j])                

    def test_fit3(self):
        from lale.lib.sklearn import MinMaxScaler, MLPClassifier, PCA
        pipeline = PCA() >> Batching(operator = MinMaxScaler() >> MLPClassifier(random_state=42), 
                                                 batch_size = 10)        
        trained = pipeline.fit(self.X_train, self.y_train)
        predictions = trained.predict(self.X_test)

    def test_no_partial_fit(self):
        pipeline = Batching(operator = NoOp() >> LogisticRegression())
        with self.assertRaises(AttributeError):
            trained = pipeline.fit(self.X_train, self.y_train)

    def test_fit4(self):
        import warnings
        warnings.filterwarnings(action="ignore")
        from lale.lib.sklearn import MinMaxScaler, MLPClassifier
        pipeline = Batching(operator = MinMaxScaler() >> MLPClassifier(random_state=42), batch_size = 112, inmemory=True)
        trained = pipeline.fit(self.X_train, self.y_train)
        predictions = trained.predict(self.X_test)
        lale_accuracy = accuracy_score(self.y_test, predictions)

        from sklearn.preprocessing import MinMaxScaler
        from sklearn.neural_network import MLPClassifier
        prep = MinMaxScaler()
        trained_prep = prep.partial_fit(self.X_train, self.y_train)
        X_transformed = trained_prep.transform(self.X_train)

        clf = MLPClassifier(random_state=42)
        import numpy as np
        trained_clf = clf.partial_fit(X_transformed, self.y_train, classes = np.unique(self.y_train))
        predictions = trained_clf.predict(trained_prep.transform(self.X_test))
        sklearn_accuracy = accuracy_score(self.y_test, predictions)

        self.assertEqual(lale_accuracy, sklearn_accuracy)        

    # TODO: Nesting doesn't work yet
    # def test_nested_pipeline(self):
    #     from lale.lib.sklearn import MinMaxScaler, MLPClassifier
    #     pipeline = Batching(operator = MinMaxScaler() >> Batching(operator = NoOp() >> MLPClassifier(random_state=42)), batch_size = 112)
    #     trained = pipeline.fit(self.X_train, self.y_train)
    #     predictions = trained.predict(self.X_test)
    #     lale_accuracy = accuracy_score(self.y_test, predictions)



from sklearn.ensemble.gradient_boosting import GradientBoostingClassifier as SKLModel
import lale.helpers
import lale.operators
from numpy import nan, inf

class GradientBoostingClassifierImpl():

    def __init__(self, loss='deviance', learning_rate=0.1, n_estimators=100, subsample=1.0, criterion='friedman_mse', min_samples_split=2, min_samples_leaf=1, min_weight_fraction_leaf=0.0, max_depth=3, min_impurity_decrease=0.0, min_impurity_split=None, init=None, random_state=None, max_features=None, verbose=0, max_leaf_nodes=None, warm_start=False, presort='auto', validation_fraction=0.1, n_iter_no_change=None, tol=0.0001):
        self._hyperparams = {
            'loss': loss,
            'learning_rate': learning_rate,
            'n_estimators': n_estimators,
            'subsample': subsample,
            'criterion': criterion,
            'min_samples_split': min_samples_split,
            'min_samples_leaf': min_samples_leaf,
            'min_weight_fraction_leaf': min_weight_fraction_leaf,
            'max_depth': max_depth,
            'min_impurity_decrease': min_impurity_decrease,
            'min_impurity_split': min_impurity_split,
            'init': init,
            'random_state': random_state,
            'max_features': max_features,
            'verbose': verbose,
            'max_leaf_nodes': max_leaf_nodes,
            'warm_start': warm_start,
            'presort': presort,
            'validation_fraction': validation_fraction,
            'n_iter_no_change': n_iter_no_change,
            'tol': tol}

    def fit(self, X, y=None):
        self._sklearn_model = SKLModel(**self._hyperparams)
        if (y is not None):
            self._sklearn_model.fit(X, y)
        else:
            self._sklearn_model.fit(X)
        return self

    def predict(self, X):
        return self._sklearn_model.predict(X)

    def predict_proba(self, X):
        return self._sklearn_model.predict_proba(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for GradientBoostingClassifier    Gradient Boosting for classification.',
    'allOf': [{
        'type': 'object',
        'required': ['loss', 'learning_rate', 'n_estimators', 'subsample', 'criterion', 'min_samples_split', 'min_samples_leaf', 'min_weight_fraction_leaf', 'max_depth', 'min_impurity_decrease', 'min_impurity_split', 'init', 'random_state', 'max_features', 'verbose', 'max_leaf_nodes', 'warm_start', 'presort', 'validation_fraction', 'n_iter_no_change', 'tol'],
        'relevantToOptimizer': ['loss', 'n_estimators', 'min_samples_split', 'min_samples_leaf', 'max_depth', 'max_features', 'presort'],
        'additionalProperties': False,
        'properties': {
            'loss': {
                'enum': ['deviance', 'exponential'],
                'default': 'deviance',
                'description': "loss function to be optimized. 'deviance' refers to"},
            'learning_rate': {
                'type': 'number',
                'default': 0.1,
                'description': 'learning rate shrinks the contribution of each tree by `learning_rate`.'},
            'n_estimators': {
                'type': 'integer',
                'minimumForOptimizer': 10,
                'maximumForOptimizer': 100,
                'distribution': 'uniform',
                'default': 100,
                'description': 'The number of boosting stages to perform. Gradient boosting'},
            'subsample': {
                'type': 'number',
                'default': 1.0,
                'description': 'The fraction of samples to be used for fitting the individual base'},
            'criterion': {
                'type': 'string',
                'default': 'friedman_mse',
                'description': 'The function to measure the quality of a split. Supported criteria'},
            'min_samples_split': {
                'anyOf': [{
                    'type': 'integer',
                    'forOptimizer': False}, {
                    'type': 'number',
                    'minimumForOptimizer': 0.01,
                    'maximumForOptimizer': 0.5,
                    'distribution': 'uniform'}],
                'default': 2,
                'description': 'The minimum number of samples required to split an internal node:'},
            'min_samples_leaf': {
                'anyOf': [{
                    'type': 'integer',
                    'forOptimizer': False}, {
                    'type': 'number',
                    'minimumForOptimizer': 0.01,
                    'maximumForOptimizer': 0.5,
                    'distribution': 'uniform'}],
                'default': 1,
                'description': 'The minimum number of samples required to be at a leaf node.'},
            'min_weight_fraction_leaf': {
                'type': 'number',
                'default': 0.0,
                'description': 'The minimum weighted fraction of the sum total of weights (of all'},
            'max_depth': {
                'type': 'integer',
                'minimumForOptimizer': 3,
                'maximumForOptimizer': 5,
                'distribution': 'uniform',
                'default': 3,
                'description': 'maximum depth of the individual regression estimators. The maximum'},
            'min_impurity_decrease': {
                'type': 'number',
                'default': 0.0,
                'description': 'A node will be split if this split induces a decrease of the impurity'},
            'min_impurity_split': {
                'anyOf': [{
                    'type': 'number'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Threshold for early stopping in tree growth. A node will split'},
            'init': {
                'XXX TODO XXX': 'estimator, optional',
                'description': 'An estimator object that is used to compute the initial',
                'enum': [None],
                'default': None},
            'random_state': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'If int, random_state is the seed used by the random number generator;'},
            'max_features': {
                'anyOf': [{
                    'type': 'integer',
                    'forOptimizer': False}, {
                    'type': 'number',
                    'minimumForOptimizer': 0.01,
                    'maximumForOptimizer': 1.0,
                    'distribution': 'uniform'}, {
                    'type': 'string',
                    'forOptimizer': False}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The number of features to consider when looking for the best split:'},
            'verbose': {
                'type': 'integer',
                'default': 0,
                'description': 'Enable verbose output. If 1 then it prints progress and performance'},
            'max_leaf_nodes': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Grow trees with ``max_leaf_nodes`` in best-first fashion.'},
            'warm_start': {
                'type': 'boolean',
                'default': False,
                'description': 'When set to ``True``, reuse the solution of the previous call to fit'},
            'presort': {
                'XXX TODO XXX': "bool or 'auto', optional (default='auto')",
                'description': 'Whether to presort the data to speed up the finding of best splits in',
                'enum': ['auto'],
                'default': 'auto'},
            'validation_fraction': {
                'type': 'number',
                'default': 0.1,
                'description': 'The proportion of training data to set aside as validation set for'},
            'n_iter_no_change': {
                'anyOf': [{
                    'type': 'integer'}, {
                    'enum': [None]}],
                'default': None,
                'description': '``n_iter_no_change`` is used to decide if early stopping will be used'},
            'tol': {
                'type': 'number',
                'default': 0.0001,
                'description': 'Tolerance for the early stopping. When the loss is not improving'},
        }}, {
        'XXX TODO XXX': 'Parameter: min_samples_leaf > only be considered if it leaves at least min_samples_leaf training samples in each of the left and right branches'}, {
        'XXX TODO XXX': 'Parameter: validation_fraction > only used if n_iter_no_change is set to an integer'}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit the gradient boosting model.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'The input samples. Internally, it will be converted to'},
        'y': {
            'type': 'array',
            'items': {
                'type': 'number'},
            'description': 'Target values (strings or integers in classification, real numbers'},
        'sample_weight': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'number'},
            }, {
                'enum': [None]}],
            'description': 'Sample weights. If None, then samples are equally weighted. Splits'},
        'monitor': {
            'anyOf': [{
                'type': 'object'}, {
                'enum': [None]}],
            'default': None,
            'description': 'The monitor is called after each iteration with the current'},
    },
}
_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict class for X.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'The input samples. Internally, it will be converted to'},
    },
}
_output_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'The predicted values.',
    'type': 'array',
    'items': {
        'type': 'number'},
}
_input_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Predict class probabilities for X.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'The input samples. Internally, it will be converted to'},
    },
}
_output_predict_proba_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'The class probabilities of the input samples. The order of the',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'},
    },
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['estimator'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output_predict': _output_predict_schema,
        'input_predict_proba': _input_predict_proba_schema,
        'output_predict_proba': _output_predict_proba_schema},
}
if (__name__ == '__main__'):
    lale.helpers.validate_is_schema(_combined_schemas)
GradientBoostingClassifier = lale.operators.make_operator(GradientBoostingClassifierImpl, _combined_schemas)




# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Iterable, List, Optional, Tuple
import itertools
import warnings
import random
import math
from collections import ChainMap

from lale.util.Visitor import Visitor
from lale.search.search_space import SearchSpace, SearchSpaceObject, SearchSpaceEnum
from lale.search.schema2search_space import schemaToSearchSpace
from lale.search.PGO import PGO
from lale.sklearn_compat import nest_all_HPparams

# To avoid import cycle, since we only realy on lale.operators for types
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from lale.operators import PlannedOperator, OperatorChoice, PlannedIndividualOp, PlannedPipeline


SearchSpaceGrid = Dict[str,SearchSpace]

def get_search_space_grids( op:'PlannedOperator', 
                            num_grids:Optional[float]=None, 
                            pgo:Optional[PGO]=None)->List[SearchSpaceGrid]:
    """ Top level function: given a lale operator, returns a list of hp grids.
    Parameters
    ----------
    op : The lale PlannedOperator
    num_grids: integer or float, optional
        if set to an integer => 1, it will determine how many parameter grids will be returned (at most)
        if set to an float between 0 and 1, it will determine what fraction should be returned
        note that setting it to 1 is treated as in integer.  To return all results, use None
    """
    all_parameters = SearchSpaceGridVisitor.run(op, pgo=pgo)
    if num_grids is None:
        return all_parameters
    else:
        if num_grids <= 0:
            warnings.warn(f"get_search_space_grids(num_grids={num_grids}) called with a non-positive value for lale_num_grids")
            return []
        if num_grids >= 1:
            samples = math.ceil(num_grids)
            if samples >= len(all_parameters):
                return all_parameters
            else:
                warnings.warn(f"get_search_space_grids(num_grids={num_grids}) sampling {math.ceil(num_grids)}/{len(all_parameters)}")
                return random.sample(all_parameters, math.ceil(num_grids))
        else:
            samples = round(len(all_parameters)*num_grids)
            warnings.warn(f"get_search_space_grids(num_grids={num_grids}) sampling {samples}/{len(all_parameters)}")
            return random.sample(all_parameters, samples)


def SearchSpaceObjectChoiceToGrid(keys:List[str], values:Tuple)->SearchSpaceGrid:
    assert len(keys) == len(values)
    return dict(zip(keys, values))


def SearchSpaceObjectectToGrid(hp:SearchSpaceObject)->List[SearchSpaceGrid]:
    return [SearchSpaceObjectChoiceToGrid(hp.keys, c) for c in hp.choices]

def searchSpaceToGrids(hp:SearchSpace)->List[SearchSpaceGrid]:
    if isinstance(hp, SearchSpaceObject):
        return SearchSpaceObjectectToGrid(hp)
    else:
        raise ValueError("Can only convert SearchSpaceObject into a GridSearchCV schema")

def schemaToSearchSpaceGrids(longName:str, 
                             name:str, 
                             schema,
                             pgo:Optional[PGO]=None)->List[SearchSpaceGrid]:
    h = schemaToSearchSpace(longName, name, schema, pgo=pgo)
    if h is None:
        return []
    grids = searchSpaceToGrids(h)
    return grids

class SearchSpaceGridVisitor(Visitor):
    pgo:Optional[PGO]

    @classmethod
    def run(cls, op:'PlannedOperator', pgo:Optional[PGO]=None):
        visitor = cls(pgo=pgo)
        accepting_op:Any = op
        return accepting_op.accept(visitor)

    def __init__(self, pgo:Optional[PGO]=None):
        super(SearchSpaceGridVisitor, self).__init__()
        self.pgo = pgo
    
    def augment_grid(self, grid:SearchSpaceGrid, hyperparams)->SearchSpaceGrid:
        if not hyperparams:
            return grid
        ret = dict(grid)
        for (k,v) in hyperparams.items():
            if k not in ret:
                ret[k] = SearchSpaceEnum([v])
        return ret

    def visitPlannedIndividualOp(self, op:'PlannedIndividualOp')->List[SearchSpaceGrid]:
        schema = op.hyperparam_schema_with_hyperparams()
        module = op._impl.__module__
        if module is None or module == str.__class__.__module__:
            long_name = op.name()
        else:
            long_name = module + '.' + op.name()
        name = op.name()
        grids = schemaToSearchSpaceGrids(long_name, name, schema, pgo=self.pgo)
        if hasattr(op, '_hyperparams'):
            hyperparams = op._hyperparams
            if hyperparams and not grids:
                grids = [{}]
            augmented_grids = [self.augment_grid(g, hyperparams) for g in grids]
            return augmented_grids
        else:
            return grids

    visitTrainableIndividualOp = visitPlannedIndividualOp
    visitTrainedIndividualOp = visitPlannedIndividualOp
    
    def visitPlannedPipeline(self, op:'PlannedPipeline')->List[SearchSpaceGrid]:

        param_grids:List[List[SearchSpaceGrid]] = [
            nest_all_HPparams(s.name(), s.accept(self)) for s in op.steps()]

        param_grids_product:Iterable[Iterable[SearchSpaceGrid]] = itertools.product(*param_grids)
        chained_grids:List[SearchSpaceGrid] = [
            dict(ChainMap(*gridline)) for gridline in param_grids_product]

        return chained_grids
    
    visitTrainablePipeline = visitPlannedPipeline
    visitTrainedPipeline = visitPlannedPipeline

    def visitOperatorChoice(self, op:'OperatorChoice')->List[SearchSpaceGrid]:

        choice_name:str = "_lale_discriminant"
        ret:List[SearchSpaceGrid] = []
        for s in op.steps():
            # if not isinstance(s, PlannedOperator):
            #     raise ValueError("This method should really be defined on PlannedOperatorChoice")
            # else:
            grids:List[SearchSpaceGrid] = s.accept(self)
            # If there are no parameters, we still need to add a choice for the discriminant
            if not grids:
                grids = [{}]
            op_name:str = s.name()
            discriminated_grids:List[SearchSpaceGrid]=[{**d, choice_name:SearchSpaceEnum([op_name])} for d in grids]
            ret.extend(discriminated_grids)
        return ret





from sklearn.preprocessing.label import LabelBinarizer as SKLModel
import lale.helpers
import lale.operators
from numpy import nan, inf

class LabelBinarizerImpl():

    def __init__(self, neg_label=0, pos_label=1, sparse_output=False):
        self._hyperparams = {
            'neg_label': neg_label,
            'pos_label': pos_label,
            'sparse_output': sparse_output}

    def fit(self, X, y=None):
        self._sklearn_model = SKLModel(**self._hyperparams)
        if (y is not None):
            self._sklearn_model.fit(X, y)
        else:
            self._sklearn_model.fit(X)
        return self

    def transform(self, X):
        return self._sklearn_model.transform(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for LabelBinarizer    Binarize labels in a one-vs-all fashion',
    'allOf': [{
        'type': 'object',
        'required': ['neg_label', 'pos_label', 'sparse_output'],
        'relevantToOptimizer': ['neg_label', 'pos_label', 'sparse_output'],
        'additionalProperties': False,
        'properties': {
            'neg_label': {
                'type': 'integer',
                'minimumForOptimizer': 0,
                'maximumForOptimizer': 1,
                'distribution': 'uniform',
                'default': 0,
                'description': 'Value with which negative labels must be encoded.'},
            'pos_label': {
                'type': 'integer',
                'minimumForOptimizer': 1,
                'maximumForOptimizer': 2,
                'distribution': 'uniform',
                'default': 1,
                'description': 'Value with which positive labels must be encoded.'},
            'sparse_output': {
                'type': 'boolean',
                'default': False,
                'description': 'True if the returned array from transform is desired to be in sparse'},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit label binarizer',
    'type': 'object',
    'properties': {
        'y': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'type': 'number'},
            }, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'Target values. The 2-d matrix should only contain 0 and 1,'},
    },
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Transform multi-class labels to binary labels',
    'type': 'object',
    'properties': {
        'y': {
            'anyOf': [{
                'type': 'array',
                'items': {
                    'XXX TODO XXX': 'item type'},
                'XXX TODO XXX': 'array or sparse matrix of shape [n_samples,] or             [n_samples, n_classes]'}, {
                'type': 'array',
                'items': {
                    'type': 'number'},
            }, {
                'type': 'array',
                'items': {
                    'type': 'array',
                    'items': {
                        'type': 'number'},
                }}],
            'description': 'Target values. The 2-d matrix should only contain 0 and 1,'},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Shape will be [n_samples, 1] for binary problems.',
    'XXX TODO XXX': 'numpy array or CSR matrix of shape [n_samples, n_classes]',
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema},
}
if (__name__ == '__main__'):
    lale.helpers.validate_is_schema(_combined_schemas)
LabelBinarizer = lale.operators.make_operator(LabelBinarizerImpl, _combined_schemas)




# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lale.helpers
import lale.operators
import numpy as np
import pandas as pd
import scipy.sparse
import jsonsubschema

class ConcatFeaturesImpl():
    """Transformer to concatenate input datasets. 

    This transformer concatenates the input datasets column-wise.

    Examples
    --------
    >>> A = [ [11, 12, 13],
              [21, 22, 23],
              [31, 32, 33] ]
    >>> B = [ [14, 15],
              [24, 25],
              [34, 35] ]
    >>> trainable_cf = ConcatFeatures()
    >>> trained_cf = trainable_cf.fit(X = [A, B])
    >>> trained_cf.transform([A, B])
        [ [11, 12, 13, 14, 15],
          [21, 22, 23, 24, 25],
          [31, 32, 33, 34, 35] ]
    """

    def __init__(self):
        pass

    def transform(self, X):
        """Transform the list of datasets to one single dataset by concatenating column-wise.
        
        Parameters
        ----------
        X : list
            List of datasets to be concatenated.
        
        Returns
        -------
        [type]
            [description]
        """
        np_datasets = []
        #Preprocess the datasets to convert them to 2-d numpy arrays
        for dataset in X:
            if isinstance(dataset, pd.DataFrame) or isinstance(dataset, pd.Series):
                np_dataset = dataset.values
            elif isinstance(dataset, scipy.sparse.csr_matrix):
                np_dataset = dataset.toarray()
            else:
                np_dataset = dataset
            if hasattr(np_dataset, 'shape'):
                if len(np_dataset.shape) == 1: #To handle numpy column vectors
                    np_dataset = np.reshape(np_dataset, (np_dataset.shape[0], 1))
            np_datasets.append(np_dataset)
            
        result = np.concatenate(np_datasets, axis=1)
        return result

    def transform_schema(self, s_X):
        min_cols, max_cols, elem_schema = 0, 0, None
        def join_schemas(s_a, s_b):
            if s_a is None:
                return s_b
            s_a = lale.helpers.dict_without(s_a, 'description')
            s_b = lale.helpers.dict_without(s_b, 'description')
            if jsonsubschema.isSubschema(s_a, s_b):
                return s_b
            if jsonsubschema.isSubschema(s_b, s_a):
                return s_a
            return jsonsubschema.joinSchemas(s_a, s_b)
        def add_ranges(min_a, max_a, min_b, max_b):
            min_ab = min_a + min_b
            if max_a == 'unbounded' or max_b == 'unbounded':
                max_ab = 'unbounded'
            else:
                max_ab = max_a + max_b
            return min_ab, max_ab
        for s_dataset in s_X['items']:
            s_rows = s_dataset['items']
            if 'type' in s_rows and 'array' == s_rows['type']:
                s_cols = s_rows['items']
                if isinstance(s_cols, dict):
                    min_c = s_rows['minItems'] if 'minItems' in s_rows else 1
                    max_c = s_rows['maxItems'] if 'maxItems' in s_rows else 'unbounded'
                    elem_schema = join_schemas(elem_schema, s_cols)
                else:
                    min_c, max_c = len(s_cols), len(s_cols)
                    for s_col in s_cols:
                        elem_schema = join_schemas(elem_schema, s_col)
                min_cols, max_cols = add_ranges(min_cols,max_cols,min_c,max_c)
            else:
                elem_schema = join_schemas(elem_schema, s_rows)
                min_cols, max_cols = add_ranges(min_cols, max_cols, 1, 1)
        s_result = {
            '$schema': 'http://json-schema.org/draft-04/schema#',
            'type': 'array',
            'items': {
                'type': 'array',
                'minItems': min_cols,
                'items': elem_schema}}
        if max_cols != 'unbounded':
            s_result['items']['maxItems'] = max_cols
        lale.helpers.validate_is_schema(s_result)
        return s_result
    
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Hyperparameter schema for the ConcatFeatures operator.\n',
    'allOf': [{
        'description': 'This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters.',
        'type': 'object',
        'additionalProperties': False,
        'relevantToOptimizer': [],
        'properties': {}}]}

_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Input data schema for training the ConcatFeatures operator. As this operator does not actually require training, this is the same as the input schema for making predictions.',
    'type': 'object',
    'required': ['X'],
    'additionalProperties': True,
    'properties': {
        'X': {
            'description': 'Outermost array dimension is over datasets.',
            'type': 'array',
            'items': {
                'description': 'Middle array dimension is over samples (aka rows).',
                'type': 'array',
                'items': {
                    'description': 'Innermost array dimension is over features (aka columns).',
                    'anyOf': [{
                        'type': 'array',
                        'items': {
                            'type': 'number'},
                    }, {
                        'type': 'number'}]}}}}}

_input_predict_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Input data schema for making predictions using the ConcatFeatures operator.',
    'type': 'object',
    'required': ['X'],
    'additionalProperties': False,
    'properties': {
        'X': {
            'description': 'Outermost array dimension is over datasets.',
            'type': 'array',
            'items': {
                'description': 'Middle array dimension is over samples (aka rows).',
                'type': 'array',
                'items': {
                    'description': 'Innermost array dimension is over features (aka columns).',
                    'anyOf': [{
                        'type': 'array',
                        'items': {
                            'type': 'number'},
                    }, {
                        'type': 'number'}]}}}}}

_output_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Output data schema for transformed data using the ConcatFeatures operator.',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'}}}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'documentation_url': 'https://lale.readthedocs.io/en/latest/modules/lale.lib.lale.concat_features.html',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_predict': _input_predict_schema,
        'output': _output_schema }}

if (__name__ == '__main__'):
    lale.helpers.validate_is_schema(_combined_schemas)

ConcatFeatures = lale.operators.make_operator(ConcatFeaturesImpl, _combined_schemas)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from spacy.lang.en import English
import numpy as np
from lale.operators import make_operator
import lale.helpers
from sklearn.utils.validation import _is_arraylike
import urllib.request
import os
import zipfile

logging.basicConfig(level=logging.INFO)


class GloveEmbeddingEncoderImpl(object):
    """
    GloveEmbeddingEncoderImpl is a module allows simple generation of sentence embeddings using
    glove word embeddings

    Parameters
    ----------
    dim: int, (default=300), dimension of word embeddings to use
    combiner: string, (default=mean), apply mean pooling or max pooling on word embeddings to generate
        sentence embedding

    References
    ----------
    R. JeffreyPennington and C. Manning. Glove: Global vectors for word representation. 2014

    """
    def __init__(self, dim=300, combiner='mean'):

        self.combiner = combiner
        self.dim = dim

        if self.combiner not in ["mean", "max"]:
            raise ValueError("Combiner must be either mean or max")

        if self.dim not in [50, 100, 200, 300]:
            raise ValueError("dim must be in 50, 100, 200, or 300")

        glove_path = os.path.join(os.path.dirname(__file__),
                                  "resources",
                                  "glove.6B.{}d.txt".format(dim))

        if not os.path.exists(glove_path):
            download_url = 'http://nlp.stanford.edu/data/glove.6B.zip'
            download_dir = os.path.join(os.path.dirname(__file__), 'resources', 'glove.6B.zip')
            if not os.path.exists(os.path.join(os.path.dirname(__file__), "resources")):
                os.mkdir(os.path.join(os.path.dirname(__file__), "resources"))
            urllib.request.urlretrieve(download_url, download_dir)
            with zipfile.ZipFile(download_dir, 'r') as zip_ref:
                zip_ref.extractall(os.path.join(os.path.dirname(__file__), "resources"))
            os.remove(download_dir)

        self.glove_dict = self._load_glove(glove_path)
        nlp = English()
        self.tokenizer = nlp.Defaults.create_tokenizer(nlp)
        print('finish initialization')

    def fit(self, X, y=None):
        return self

    def _load_glove(self, glove_path):
        glove_dict = dict()
        with open(glove_path, 'r', encoding='utf-8') as file_handle:
            for line in file_handle:
                line = line.split()
                glove_dict[line[0]] = np.array(line[1:], dtype=np.float32)
        return glove_dict

    def transform(self, X):
        if not _is_arraylike(X):
            raise TypeError("X is not iterable")

        transformed_X = list()
        for text in X:
            temp_vec = list()
            for token in self.tokenizer(text.lower()):
                temp_vec.append(self.glove_dict.get(token, self.glove_dict['unk']))

            if self.combiner == 'mean':
                sentence_vec = np.mean(temp_vec, axis=0)
            else:
                sentence_vec = np.amax(temp_vec, axis=0)

            transformed_X.append(sentence_vec)
        return transformed_X

_input_schema_fit = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Input data schema for training.',
    'type': 'object',
    'required': ['X'],
    'additionalProperties': False,
    'properties': {
        'X': {
            'description': 'Input Text',
            'anyOf': [
                {'type': 'array',
                 'items': {'type': 'string'}},
                {'type': 'array',
                 'items': {
                     'type': 'array', 'minItems': 1, 'maxItems': 1,
                     'items': {'type': 'string'}}}]}
    }}

_input_schema_predict = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Input data schema for training.',
    'type': 'object',
    'required': ['X'],
    'additionalProperties': False,
    'properties': {
        'X': {
            'description': 'Input Text',
            'anyOf': [
                {'type': 'array',
                 'items': {'type': 'string'}},
                {'type': 'array',
                 'items': {
                     'type': 'array', 'minItems': 1, 'maxItems': 1,
                     'items': {'type': 'string'}}}]}
    }}

_output_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Output data schema for transformed data.',
    'type': 'array',
    'items': {'type': 'array', 'items': {'type': 'number'}}}

_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Hyperparameter schema.',
    'allOf': [
        {'description':
             'This first sub-object lists all constructor arguments with their '
             'types, one at a time, omitting cross-argument constraints.',
         'type': 'object',
         'additionalProperties': False,
         'relevantToOptimizer': [],
         'properties': {
             'combiner': {
                'enum': ['mean', 'max'],
                'default': 'mean'},
                'dim':{'enum': [50, 100, 200, 300],
                       'default': 300}
         }}]}

_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters for a transformer for'
                   ' a text data transformer based on pre-trained glove embedding',
    'type': 'object',
    'tags': {
        'pre': ['text'],
        'op': ['transformer', '~interpretable'],
        'post': ['embedding']},
    'properties': {
        'input_fit': _input_schema_fit,
        'input_predict': _input_schema_predict,
        'output': _output_schema,
        'hyperparams': _hyperparams_schema}}


if __name__ == "__main__":
    lale.helpers.validate_is_schema(_combined_schemas)

GloveEmbeddingEncoder = make_operator(GloveEmbeddingEncoderImpl(), _combined_schemas)


# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import warnings
import random
import jsonschema
import sys
import lale.operators as Ops
from lale.lib.lale import ConcatFeatures
from lale.lib.lale import NoOp
from lale.lib.sklearn import KNeighborsClassifier
from lale.lib.sklearn import LinearSVC
from lale.lib.sklearn import LogisticRegression
from lale.lib.sklearn import MinMaxScaler
from lale.lib.sklearn import MLPClassifier
from lale.lib.sklearn import Nystroem
from lale.lib.sklearn import OneHotEncoder
from lale.lib.sklearn import PCA
from lale.lib.sklearn import TfidfVectorizer
from lale.lib.sklearn import MultinomialNB
from lale.lib.sklearn import SimpleImputer
from lale.lib.sklearn import SVC
from lale.lib.xgboost import XGBClassifier
from lale.lib.sklearn import PassiveAggressiveClassifier
from lale.lib.sklearn import StandardScaler
from lale.lib.sklearn import FeatureAgglomeration
from typing import List

import sklearn.datasets

from lale.sklearn_compat import make_sklearn_compat
from lale.search.GridSearchCV import LaleGridSearchCV, get_grid_search_parameter_grids
from lale.search.SMAC import get_smac_space, lale_trainable_op_from_config
from lale.search.op2hp import hyperopt_search_space

@unittest.skip("Skipping here because travis-ci fails to allocate memory. This runs on internal travis.")
class TestResNet50(unittest.TestCase):
    def test_init_fit_predict(self):
        import torchvision.datasets as datasets
        import torchvision.transforms as transforms
        from lale.lib.pytorch import ResNet50

        transform = transforms.Compose([transforms.ToTensor()])

        data_train = datasets.FakeData(size = 50, num_classes=2 , transform = transform)#, target_transform = transform)
        clf = ResNet50(num_classes=2,num_epochs = 1)
        clf.fit(data_train)
        predicted = clf.predict(data_train)




# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import contextlib
import json
import jsonschema
import jsonsubschema
import numpy as np
import pandas as pd
import os
import re
import sys
import time
import traceback
import urllib
import warnings
import yaml
import scipy.sparse
import importlib
import graphviz
from sklearn.model_selection import StratifiedKFold
from sklearn.metrics import accuracy_score, log_loss
from sklearn.utils.metaestimators import _safe_split
from lale.util.numpy_to_torch_dataset import NumpyTorchDataset
from lale.util.hdf5_to_torch_dataset import HDF5TorchDataset
from torch.utils.data import DataLoader
import copy
import logging
import importlib
import inspect
import pkgutil
import torch
import h5py
logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger(__name__)

class assert_raises:
    def __init__(self, expected_exc_type):
        self.__expected_exc_type = expected_exc_type

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        assert exc_type is self.__expected_exc_type
        print_yaml('error', str(exc_value), file=sys.stderr)
        return True

class assert_warns:
    def __init__(self, expected_exc_type):
        self.__expected_exc_type = expected_exc_type

    def __enter__(self):
        warnings.filterwarnings("error", category=self.__expected_exc_type)

    def __exit__(self, exc_type, exc_value, traceback):
        warnings.resetwarnings()
        assert exc_type is self.__expected_exc_type
        print_yaml('error', str(exc_value), file=sys.stderr)
        return True

def assignee_name(level=1):
    tb = traceback.extract_stack()
    file_name, line_number, function_name, text = tb[-(level+2)]
    tree = ast.parse(text, file_name)
    assert isinstance(tree, ast.Module)
    if len(tree.body) == 1 and isinstance(tree.body[0], ast.Assign):
        lhs = tree.body[0].targets
        if len(lhs) == 1 and isinstance(lhs[0], ast.Name):
            return lhs[0].id
    return None

def data_to_json(data, subsample_array = True):
    if type(data) is tuple:
        # convert to list
        return [data_to_json(elem, subsample_array) for elem in data]
    if type(data) is list:
        return [data_to_json(elem, subsample_array) for elem in data]
    elif type(data) is dict:
        return {key: data_to_json(data[key], subsample_array) for key in data}
    elif isinstance(data, np.ndarray):
        return ndarray_to_json(data, subsample_array)
    elif type(data) is scipy.sparse.csr_matrix:
        return ndarray_to_json(data.toarray(), subsample_array)
    elif isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
        np_array = data.values
        return ndarray_to_json(np_array, subsample_array)
    elif isinstance(data, torch.Tensor):
        np_array = data.numpy()
        return ndarray_to_json(np_array, subsample_array)
    else:
        return data

def dict_without(orig_dict, key):
    return {k: orig_dict[k] for k in orig_dict if k != key}

def load_yaml(dir_name, file_name, meta_dir=True):
    current_dir = os.path.dirname(__file__)
    parent_dir = os.path.dirname(current_dir)
    if meta_dir:
        op_dir = os.path.join(parent_dir, 'meta_data', dir_name)
    else:
        op_dir = os.path.join(parent_dir, dir_name)
    schema_file = os.path.join(op_dir, file_name)
    with open(schema_file, 'r') as f:
        result = yaml.load(f)
    return result

def ndarray_to_json(arr, subsample_array=True):
    #sample 10 rows and no limit on columns
    if subsample_array:
        num_subsamples = [10, np.iinfo(np.int).max, np.iinfo(np.int).max]
    else:
        num_subsamples = [np.iinfo(np.int).max, np.iinfo(np.int).max, np.iinfo(np.int).max]
    def subarray_to_json(indices):
        if len(indices) == len(arr.shape):
            if isinstance(arr[indices], float) or isinstance(arr[indices], int)\
                or isinstance(arr[indices], str):
                return arr[indices]
            elif arr.dtype == np.float64 or arr.dtype == np.float32:
                return float(arr[indices])
            elif arr.dtype == np.int64:
                return int(arr[indices])
            elif arr.dtype == np.bool:
                return bool(arr[indices])
            elif arr.dtype.kind == 'U':
                return str(arr[indices])
            elif arr.dtype.kind == 'O':
                return str(arr[indices])
            else:
                raise ValueError(f'Unexpected dtype {arr.dtype}, '
                                 f'kind {arr.dtype.kind}, '
                                 f'type {type(arr[indices])}.')
        else:
            assert len(indices) < len(arr.shape)
            return [subarray_to_json(indices + (i,))
                    for i in range(min(num_subsamples[len(indices)], arr.shape[len(indices)]))]
    return subarray_to_json(())

def print_yaml(what, doc, file=sys.stdout):
    print(yaml.dump({what: doc}).strip(), file=file)

def validate_schema(value, schema, subsample_array=True):
    json_value = data_to_json(value, subsample_array)
    jsonschema.validate(json_value, schema)

JSON_META_SCHEMA_URL = 'http://json-schema.org/draft-04/schema#'
_JSON_META_SCHEMA = None

def json_meta_schema():
    global _JSON_META_SCHEMA
    if _JSON_META_SCHEMA is None:
        url = JSON_META_SCHEMA_URL
        with contextlib.closing(urllib.request.urlopen(url)) as f:
            _JSON_META_SCHEMA = json.load(f)
    return _JSON_META_SCHEMA

def validate_is_schema(value):
    if '$schema' in value:
        assert value['$schema'] == JSON_META_SCHEMA_URL
    jsonschema.validate(value, json_meta_schema())

def is_schema(value):
    if isinstance(value, dict):
        try:
            jsonschema.validate(value, json_meta_schema())
        except:
            return False
        return True
    return False

class SubschemaError(Exception):
    def __init__(self, sub, sup, sub_name='sub', sup_name='super'):
        self.sub = sub
        self.sup = sup
        self.sub_name = sub_name
        self.sup_name = sup_name

    def __str__(self):
        summary = f'expected {self.sub_name} <: {self.sup_name}'
        import pprint
        sub = pprint.pformat(self.sub, width=70, compact=True)
        sup = pprint.pformat(self.sup, width=70, compact=True)
        details = f'\n{self.sub_name} = \\\n{sub}\n{self.sup_name} = \\\n{sup}'
        return summary + details

def validate_subschema(sub, sup, sub_name='sub', sup_name='super'):
    if not jsonsubschema.isSubschema(sub, sup):
        raise SubschemaError(sub, sup, sub_name, sup_name)

def cross_val_score_track_trials(estimator, X, y=None, scoring=accuracy_score, cv=5):
    """
    Use the given estimator to perform fit and predict for splits defined by 'cv' and compute the given score on 
    each of the splits.

    :param estimator: A valid sklearn_wrapper estimator
    :param X, y: Valid data and target values that work with the estimator
    :param scoring: a scorer object from sklearn.metrics (https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics)
             Default value is accuracy_score.
    :param cv: an integer or an object that has a split function as a generator yielding (train, test) splits as arrays of indices.
        Integer value is used as number of folds in sklearn.model_selection.StratifiedKFold, default is 5.
        Note that any of the iterators from https://scikit-learn.org/stable/modules/cross_validation.html#cross-validation-iterators can be used here.

    :return: cv_results: a list of scores corresponding to each cross validation fold
    """
    if isinstance(cv, int):
        cv = StratifiedKFold(cv)
    
    cv_results = []
    log_loss_results = []
    time_results = []
    for train, test in cv.split(X, y):
        X_train, y_train = _safe_split(estimator, X, y, train)
        X_test, y_test = _safe_split(estimator, X, y, test, train)
        start = time.time()
        trained_estimator = estimator.fit(X_train, y_train)
        predicted_values = trained_estimator.predict(X_test)
        execution_time = time.time() - start
        # not all estimators have predict probability
        try:
            y_pred_proba = trained_estimator.predict_proba(X_test)
            logloss = log_loss(y_true=y_test, y_pred=y_pred_proba)
            log_loss_results.append(logloss)
        except BaseException:
            logger.debug("Warning, log loss cannot be computed")
        cv_results.append(scoring(y_test, predicted_values))
        time_results.append(execution_time)

    return np.array(cv_results).mean(), np.array(log_loss_results).mean(), np.array(execution_time).mean()


def cross_val_score(estimator, X, y=None, scoring=accuracy_score, cv=5):
    """
    Use the given estimator to perform fit and predict for splits defined by 'cv' and compute the given score on
    each of the splits.
    :param estimator: A valid sklearn_wrapper estimator
    :param X, y: Valid data and target values that work with the estimator
    :param scoring: a scorer object from sklearn.metrics (https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics)
             Default value is accuracy_score.
    :param cv: an integer or an object that has a split function as a generator yielding (train, test) splits as arrays of indices.
        Integer value is used as number of folds in sklearn.model_selection.StratifiedKFold, default is 5.
        Note that any of the iterators from https://scikit-learn.org/stable/modules/cross_validation.html#cross-validation-iterators can be used here.
    :return: cv_results: a list of scores corresponding to each cross validation fold
    """
    if isinstance(cv, int):
        cv = StratifiedKFold(cv)

    cv_results = []
    for train, test in cv.split(X, y):
        X_train, y_train = _safe_split(estimator, X, y, train)
        X_test, y_test = _safe_split(estimator, X, y, test, train)
        trained_estimator = estimator.fit(X_train, y_train)
        predicted_values = trained_estimator.predict(X_test)
        cv_results.append(scoring(y_test, predicted_values))

    return cv_results

def create_operator_using_reflection(class_name, operator_name, param_dict):
    instance = None
    if class_name is not None:
        class_name_parts = class_name.split(".")
        assert(len(class_name_parts)) >1, "The class name needs to be fully qualified, i.e. module name + class name"
        module_name = ".".join(class_name_parts[0:-1])
        class_name = class_name_parts[-1]

        module = importlib.import_module(module_name)

        class_ = getattr(module, class_name)
        
        if param_dict is None:
            instance = class_.create(name=operator_name)
        else:
            instance = class_.create(name=operator_name, kwargs = param_dict)
    return instance

def create_individual_op_using_reflection(class_name, operator_name, param_dict):
    instance = None
    if class_name is not None:
        class_name_parts = class_name.split(".")
        assert(len(class_name_parts)) >1, "The class name needs to be fully qualified, i.e. module name + class name"
        module_name = ".".join(class_name_parts[0:-1])
        class_name = class_name_parts[-1]

        module = importlib.import_module(module_name)
        class_ = getattr(module, class_name)
        
        if param_dict is None:
            instance = class_()
        else:
            instance = class_(**param_dict)
    return instance

def to_graphviz(lale_operator, **dot_graph_attr):
    from lale.operators import Operator, Pipeline
    from lale.pretty_print import hyperparams_to_string
    if not isinstance(lale_operator, Operator):
        raise ValueError("The input to to_graphviz needs to be a valid LALE operator.")
    jsn = lale_operator.to_json()
    dot = graphviz.Digraph()
    dot.attr('graph', {**dot_graph_attr, 'rankdir': 'LR'})
    dot.attr('node', fontsize='11', margin='0.06,0.03')
    if isinstance(lale_operator, Pipeline):
        nodes = jsn['steps']
    else:
        nodes = [jsn]
    #figure out what things were called in the caller
    cls2sym, name2cls = {}, {}
    import inspect
    symtab = inspect.stack()[1][0].f_globals
    for sym, val in symtab.items():
        if isinstance(val, Operator):
            cls = val.class_name()
            cls2sym[cls] = '?' if (cls in cls2sym) else sym
    cls2sym = {cls: cls2sym[cls] for cls in cls2sym if cls2sym[cls] != '?'}
    for node in nodes:
        name, cls = node['operator'], node['class']
        if '|' in name:
            for step in node['steps']:
                name2cls[step['operator']] = step['class']
        else:
            name2cls[name] = cls
    def name2sym(name):
        if '|' in name:
            step_syms = [name2sym(n.strip()) for n in name.split(sep='|')]
            return ' | '.join(step_syms)
        if name in name2cls:
            cls = name2cls[name]
            if cls in cls2sym:
                return cls2sym[cls]
        return name
    def name2label(name):
        sym = name2sym(name)
        if sym.find('|') == -1:
            l1 = re.sub('(.)([A-Z][a-z]+)', r'\1-\n\2', sym)
            l2 = re.sub('([a-z0-9])([A-Z])', r'\1-\n\2', l1)
            label = re.sub(r'([^_\n-]_)([^_\n-])', r'\1-\n\2', l2)
        else:
            label = sym
        return label
    #do the actual visualizations
    try:
        edges = jsn['edges']
    except KeyError:
        edges = []
    def make_tooltip(node):
        sym = name2sym(node['operator'])
        if 'hyperparams' in node:
            hps = node['hyperparams']
            if hps:
                result = f'{sym}({hyperparams_to_string(hps)})'
                if len(result) > 255: #too long for graphviz
                    result = result[:252] + '...'
                return result
        return sym
    for i, node in enumerate(nodes):
        state2color = {
            'trained': 'white',
            'trainable': 'lightskyblue1',
            'planned': 'skyblue2'}
        attrs = {
            'style' :'filled',
            'fillcolor': state2color[node['state']],
            'tooltip': make_tooltip(node)}
        if 'documentation_url' in node:
            attrs = {**attrs, 'URL': node['documentation_url']}
        dot.node(str(i), name2label(node['operator']), **attrs)
    for edge in edges:
        dot.edge(str(edge[0]), str(edge[1]))
    return dot

def println_pos(message, out_file=sys.stdout):
    tb = traceback.extract_stack()[-2]
    match = re.search(r'<ipython-input-([0-9]+)-', tb[0])
    if match:
        pos = 'notebook cell [{}] line {}'.format(match[1], tb[1])
    else:
        pos = '{}:{}'.format(tb[0], tb[1])
    strtime = time.strftime('%Y-%m-%d_%H-%M-%S')
    to_log = '{}: {} {}'.format(pos, strtime, message)
    print(to_log, file=out_file)
    if match:
        os.system('echo {}'.format(to_log))

def create_instance_from_hyperopt_search_space(lale_object, hyperparams):
    '''
    Hyperparams is a n-tuple of dictionaries of hyper-parameters, each
    dictionary corresponds to an operator in the pipeline
    '''
    #lale_object can either be an individual operator, a pipeline or an operatorchoice
    #Validate that the number of elements in the n-tuple is the same
    #as the number of steps in the current pipeline

    from lale.operators import IndividualOp
    from lale.operators import Pipeline
    from lale.operators import TrainablePipeline
    from lale.operators import OperatorChoice
    if isinstance(lale_object, IndividualOp):
        return lale_object(**dict_without(hyperparams, 'name'))
    elif isinstance(lale_object, Pipeline):
        if len(hyperparams) != len(lale_object.steps()):
            raise ValueError('The number of steps in the hyper-parameter space does not match the number of steps in the pipeline.')
        op_instances = []
        edges = lale_object.edges()
        #op_map:Dict[PlannedOpType, TrainableOperator] = {}
        op_map = {}
        for op_index in range(len(hyperparams)):
            state_params = hyperparams[op_index]
            #TODO: Should ideally check if the class_name is the same as the class name of the op from self.operators() at op_index
            op_instance = create_instance_from_hyperopt_search_space(lale_object.steps()[op_index], state_params)
            op_instances.append(op_instance)
            orig_op = lale_object._steps[op_index]
            op_map[orig_op] = op_instance

        #trainable_edges:List[Tuple[TrainableOperator, TrainableOperator]]
        try:
            trainable_edges = [(op_map[x], op_map[y]) for (x, y) in edges]
        except KeyError as e:
            raise ValueError("An edge was found with an endpoint that is not a step (" + str(e) + ")")

        return TrainablePipeline(op_instances, trainable_edges, ordered=True)
    elif isinstance(lale_object, OperatorChoice):
        #Hyperopt search space for an OperatorChoice is generated as a dictionary with a single element
        #corresponding to the choice made, the only key is the index of the step and the value is 
        #the params corresponding to that step.
        step_index, hyperparams = list(hyperparams.items())[0]
        step_object = lale_object.steps()[step_index]
        return create_instance_from_hyperopt_search_space(step_object, hyperparams)
        
def import_from_sklearn_pipeline(sklearn_pipeline):
    #For all pipeline steps, identify equivalent lale wrappers if present,
    #if not, call make operator on sklearn classes and create a lale pipeline.

    def get_equivalent_lale_op(sklearn_obj):
        module_name = "lale.lib.sklearn"
        from sklearn.base import clone
        from lale.operators import make_operator

        class_name = sklearn_obj.__class__.__name__
        module = importlib.import_module(module_name)
        try:
            class_ = getattr(module, class_name)
        except AttributeError:
            class_ = make_operator(sklearn_obj, name=class_name)
        class_ = class_(**sklearn_obj.get_params())
        class_._impl._sklearn_model =  copy.deepcopy(sklearn_obj)
        return class_         

    from sklearn.pipeline import FeatureUnion, Pipeline
    from sklearn.base import BaseEstimator
    from lale.operators import make_pipeline, make_union
    
    if isinstance(sklearn_pipeline, Pipeline):
        nested_pipeline_steps = sklearn_pipeline.named_steps.values()
        nested_pipeline_lale_objects = [import_from_sklearn_pipeline(nested_pipeline_step) for nested_pipeline_step in nested_pipeline_steps]
        lale_op_obj = make_pipeline(*nested_pipeline_lale_objects)
    elif isinstance(sklearn_pipeline, FeatureUnion):
        transformer_list = sklearn_pipeline.transformer_list
        concat_predecessors = [import_from_sklearn_pipeline(transformer[1]) for transformer in transformer_list]
        lale_op_obj = make_union(*concat_predecessors)
    else:
        lale_op_obj = get_equivalent_lale_op(sklearn_pipeline)
    return lale_op_obj
        
def get_hyperparam_names(op):
    import inspect
    if op._impl.__module__.startswith('lale'):
        hp_schema = op.hyperparam_schema()
        params = next(iter(hp_schema.get('allOf', []))).get('properties', {})
        return list(params.keys())
    else:
        return inspect.getargspec(op._impl.__class__.__init__).args
                
def validate_method(op, m):
    if op._impl.__module__.startswith('lale'):
        assert (m in op._schemas['properties'].keys())
    else:
        a = ''
        if m.startswith('input_'):
            a = m[len('input_'):]
        elif m.startswith('output_'):
            a = m[len('output_'):]
        if a:
            assert (hasattr(op._impl, a))

def caml_to_snake(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

def get_lib_schema(impl):
    module_name = impl.__module__.split('.')[0]
    class_name = caml_to_snake(impl.__class__.__name__)
    try:
        lib_name = '.'.join(['lale.lib', module_name, class_name])
        m = importlib.import_module(lib_name)
        return m._combined_schemas
    except (ModuleNotFoundError, AttributeError):
        try:
            lib_name = '.'.join(['lale.lib.autogen', class_name])
            m = importlib.import_module(lib_name)
            return m._combined_schemas
        except:
            return None

def signature_to_schema(sig):
    sig_schema = {'type': 'object', 'properties': {}}
    for name, param in sig.parameters.items():
        ignored_kinds = [inspect.Parameter.VAR_POSITIONAL,
                         inspect.Parameter.VAR_KEYWORD]
        if param.kind not in ignored_kinds:
            param_schema = {}
            if param.default != inspect.Parameter.empty:
                param_schema['default'] = param.default
            sig_schema['properties'][name] = param_schema
    return sig_schema

def get_default_schema(impl):
    if hasattr(impl, '__init__'):
        sig = inspect.signature(impl.__init__)
        arg_schemas = signature_to_schema(sig)
    else:
        arg_schemas = {'type': 'object', 'properties': {}}
    return {
        '$schema': 'http://json-schema.org/draft-04/schema#',
        'description':
        'Combined schema for expected data and hyperparameters.',
        'type': 'object',
        'properties': {
            'input_fit': {},
            'input_predict': {},
            'output': {},
            'hyperparams': {
                'allOf': [arg_schemas]}}}
    
logger = logging.getLogger(__name__)

def wrap_imported_operators():
    import lale.lib
    from lale.operators import Operator, make_operator
    calling_frame = inspect.stack()[1][0]
    symtab = calling_frame.f_globals
    lib_modules = [p.name for p in pkgutil.iter_modules(lale.lib.__path__)]
    for name, impl in symtab.items():
        if inspect.isclass(impl) and not isinstance(impl, Operator):
            module = impl.__module__.split('.')[0]
            looks_like_op = hasattr(impl, 'fit') and (
                hasattr(impl, 'predict') or hasattr(impl, 'transform'))
            if module in lib_modules or looks_like_op:
                logger.info(f'Lale:Wrapped operator:{name}')
                symtab[name] = make_operator(impl=impl, name=name)

class val_wrapper():
    """This is used to wrap values that cause problems for hyper-optimizer backends
    lale will unwrap these when given them as the value of a hyper-parameter"""

    def __init__(self, base):
        self._base = base

    def unwrap_self(self):
        return self._base

    @classmethod
    def unwrap(cls, obj):
        if isinstance(obj, cls):
            return cls.unwrap(obj.unwrap_self())
        else:
            return obj

def append_batch(data, batch_data):
    if data is None:
        return batch_data
    elif isinstance(data, np.ndarray):
        if isinstance(batch_data, np.ndarray):
            if len(data.shape) == 1 and len(batch_data.shape) == 1:
                return np.concatenate([data, batch_data])
            else:
                return np.vstack((data, batch_data))
    elif isinstance(data, tuple):
        X, y = data
        if isinstance(batch_data, tuple):
            batch_X, batch_y = batch_data
            X = append_batch(X, batch_X)
            y = append_batch(y, batch_y)
            return X, y
    elif isinstance(data, torch.Tensor):
        if isinstance(batch_data, torch.Tensor):
            return torch.cat((data, batch_data))
    elif isinstance(data, h5py.File):
        if isinstance(batch_data, tuple):
            batch_X, batch_y = batch_data
            
    #TODO:Handle dataframes

def create_data_loader(X, y = None, batch_size = 1):
    if isinstance(X, pd.DataFrame):
        X = X.to_numpy()
        if isinstance(y, pd.Series):
            y = y.to_numpy()
    if isinstance(X, np.ndarray):
        dataset = NumpyTorchDataset(X, y)
    if isinstance(X, str):#Assume that this is path to hdf5 file
        dataset = HDF5TorchDataset(X)
    return DataLoader(dataset, batch_size=batch_size)

def write_batch_output_to_file(file_obj, file_path, total_len, batch_idx, batch_X, batch_y, batch_out_X, batch_out_y):
    if file_obj is None and file_path is None:
        raise ValueError("Only one of the file object or file path can be None.")
    if file_obj is None:
        file_obj = h5py.File(file_path, 'w')
        #estimate the size of the dataset based on the first batch output size
        transform_ratio = int(len(batch_out_X)/len(batch_X))
        if len(batch_out_X.shape) == 1:
            h5_data_shape = (transform_ratio*total_len, )
        if len(batch_out_X.shape) == 2:
            h5_data_shape = (transform_ratio*total_len, batch_out_X.shape[1])
        elif len(batch_out_X.shape) == 3:
            h5_data_shape = (transform_ratio*total_len, batch_out_X.shape[1], batch_out_X.shape[2])
        dataset = file_obj.create_dataset(name='X', shape=h5_data_shape, chunks=True, compression="gzip")
        if batch_out_y is None and batch_y is not None:
            batch_out_y = batch_y
        if batch_out_y is not None:
            if len(batch_out_y.shape) == 1:
                h5_labels_shape = (transform_ratio*total_len, )
            elif len(batch_out_y.shape) == 2:
                h5_labels_shape = (transform_ratio*total_len, batch_out_y.shape[1])
            dataset = file_obj.create_dataset(name='y', shape=h5_labels_shape, chunks=True, compression="gzip")
    dataset = file_obj['X']
    dataset[batch_idx*len(batch_out_X):(batch_idx+1)*len(batch_out_X)] = batch_out_X
    if batch_out_y is not None or batch_y is not None:
        labels = file_obj['y']
        if batch_out_y is not None:
            labels[batch_idx*len(batch_out_y):(batch_idx+1)*len(batch_out_y)] = batch_out_y
        else:
            labels[batch_idx*len(batch_y):(batch_idx+1)*len(batch_y)] = batch_y
    return file_obj


# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from lale.operators import make_pipeline
from lale.operators import TrainablePipeline, TrainedPipeline
from lale.helpers import import_from_sklearn_pipeline

import pickle

from lale.lib.lale import ConcatFeatures
from lale.lib.lale import NoOp
from lale.lib.lale import Batching
from lale.lib.sklearn import KNeighborsClassifier
from lale.lib.sklearn import LinearSVC
from lale.lib.sklearn import LogisticRegression
from lale.lib.sklearn import MinMaxScaler
from lale.lib.sklearn import MLPClassifier
from lale.lib.sklearn import Nystroem
from lale.lib.sklearn import OneHotEncoder
from lale.lib.sklearn import PCA
from lale.lib.sklearn import TfidfVectorizer
from lale.lib.sklearn import MultinomialNB
from lale.lib.sklearn import SimpleImputer
from lale.lib.sklearn import SVC
from lale.lib.xgboost import XGBClassifier
from lale.lib.sklearn import PassiveAggressiveClassifier
from lale.lib.sklearn import StandardScaler
from lale.lib.sklearn import FeatureAgglomeration
from lale.lib.autogen import SGDClassifier
from lale.search.GridSearchCV import LaleGridSearchCV, get_grid_search_parameter_grids
from sklearn.metrics import accuracy_score

class TestImportExport(unittest.TestCase):
    def setUp(self):
        from sklearn.datasets import load_iris
        from sklearn.model_selection import train_test_split
        data = load_iris()
        X, y = data.data, data.target
        self.X_train, self.X_test, self.y_train, self.y_test =  train_test_split(X, y)    

    def assert_equal_predictions(self, pipeline1, pipeline2):
        trained = pipeline1.fit(self.X_train, self.y_train)
        predictions1 = trained.predict(self.X_test)

        trained = pipeline2.fit(self.X_train, self.y_train)
        predictions2 = trained.predict(self.X_test)
        [self.assertEqual(p1, predictions2[i]) for i, p1 in enumerate(predictions1)]

    def test_import_from_sklearn_pipeline(self):
        from sklearn.feature_selection import SelectKBest
        from sklearn.feature_selection import f_regression        
        from sklearn import svm
        from sklearn.pipeline import Pipeline
        anova_filter = SelectKBest(f_regression, k=3)
        clf = svm.SVC(kernel='linear')        
        sklearn_pipeline = Pipeline([('anova', anova_filter), ('svc', clf)])  
        lale_pipeline = import_from_sklearn_pipeline(sklearn_pipeline)
        for i, pipeline_step in enumerate(sklearn_pipeline.named_steps):
            sklearn_step_params = sklearn_pipeline.named_steps[pipeline_step].get_params()
            lale_sklearn_params = lale_pipeline.steps()[i]._impl._sklearn_model.get_params()
            self.assertEqual(sklearn_step_params, lale_sklearn_params)
        self.assert_equal_predictions(sklearn_pipeline, lale_pipeline)


    def test_import_from_sklearn_pipeline1(self):
        from sklearn.decomposition import PCA
        from sklearn.neighbors import KNeighborsClassifier
        from sklearn.pipeline import make_pipeline
        sklearn_pipeline = make_pipeline(PCA(n_components=3), KNeighborsClassifier())
        lale_pipeline = import_from_sklearn_pipeline(sklearn_pipeline)
        for i, pipeline_step in enumerate(sklearn_pipeline.named_steps):
            sklearn_step_params = sklearn_pipeline.named_steps[pipeline_step].get_params()
            lale_sklearn_params = lale_pipeline.steps()[i]._impl._sklearn_model.get_params()
            self.assertEqual(sklearn_step_params, lale_sklearn_params)
        self.assert_equal_predictions(sklearn_pipeline, lale_pipeline)

    def test_import_from_sklearn_pipeline_feature_union(self):
        from sklearn.pipeline import FeatureUnion        
        from sklearn.decomposition import PCA
        from sklearn.kernel_approximation import Nystroem
        from sklearn.neighbors import KNeighborsClassifier
        from sklearn.pipeline import make_pipeline
        union = FeatureUnion([("pca", PCA(n_components=1)), ("nys", Nystroem(n_components=2, random_state=42))])        
        sklearn_pipeline = make_pipeline(union, KNeighborsClassifier())
        lale_pipeline = import_from_sklearn_pipeline(sklearn_pipeline)
        self.assertEqual(len(lale_pipeline.edges()), 3)
        from lale.lib.sklearn.pca import PCAImpl
        from lale.lib.sklearn.nystroem import NystroemImpl
        from lale.lib.lale.concat_features import ConcatFeaturesImpl
        from lale.lib.sklearn.k_neighbors_classifier import KNeighborsClassifierImpl
        self.assertIsInstance(lale_pipeline.edges()[0][0]._impl, PCAImpl)
        self.assertIsInstance(lale_pipeline.edges()[0][1]._impl, ConcatFeaturesImpl)
        self.assertIsInstance(lale_pipeline.edges()[1][0]._impl, NystroemImpl)
        self.assertIsInstance(lale_pipeline.edges()[1][1]._impl, ConcatFeaturesImpl)
        self.assertIsInstance(lale_pipeline.edges()[2][0]._impl, ConcatFeaturesImpl)
        self.assertIsInstance(lale_pipeline.edges()[2][1]._impl, KNeighborsClassifierImpl)
        self.assert_equal_predictions(sklearn_pipeline, lale_pipeline)

    def test_import_from_sklearn_pipeline_nested_pipeline(self):
        from sklearn.pipeline import FeatureUnion, make_pipeline       
        from sklearn.decomposition import PCA
        from sklearn.kernel_approximation import Nystroem
        from sklearn.feature_selection import SelectKBest
        from sklearn.neighbors import KNeighborsClassifier
        from sklearn.pipeline import make_pipeline
        union = FeatureUnion([("selectkbest_pca", make_pipeline(SelectKBest(k=3), PCA(n_components=1))), ("nys", Nystroem(n_components=2, random_state=42))])        
        sklearn_pipeline = make_pipeline(union, KNeighborsClassifier())
        lale_pipeline = import_from_sklearn_pipeline(sklearn_pipeline)
        self.assertEqual(len(lale_pipeline.edges()), 4)
        from lale.lib.sklearn.pca import PCAImpl
        from lale.lib.sklearn.nystroem import NystroemImpl
        from lale.lib.lale.concat_features import ConcatFeaturesImpl
        from lale.lib.sklearn.k_neighbors_classifier import KNeighborsClassifierImpl
        #These assertions assume topological sort
        self.assertIsInstance(lale_pipeline.edges()[0][0]._impl, SelectKBest)
        self.assertIsInstance(lale_pipeline.edges()[0][1]._impl, PCAImpl)
        self.assertIsInstance(lale_pipeline.edges()[1][0]._impl, PCAImpl)
        self.assertIsInstance(lale_pipeline.edges()[1][1]._impl, ConcatFeaturesImpl)
        self.assertIsInstance(lale_pipeline.edges()[2][0]._impl, NystroemImpl)
        self.assertIsInstance(lale_pipeline.edges()[2][1]._impl, ConcatFeaturesImpl)
        self.assertIsInstance(lale_pipeline.edges()[3][0]._impl, ConcatFeaturesImpl)
        self.assertIsInstance(lale_pipeline.edges()[3][1]._impl, KNeighborsClassifierImpl)
        self.assert_equal_predictions(sklearn_pipeline, lale_pipeline)

    def test_import_from_sklearn_pipeline_nested_pipeline1(self):
        from sklearn.pipeline import FeatureUnion, make_pipeline       
        from sklearn.decomposition import PCA
        from sklearn.kernel_approximation import Nystroem
        from sklearn.feature_selection import SelectKBest
        from sklearn.neighbors import KNeighborsClassifier
        from sklearn.pipeline import make_pipeline
        union = FeatureUnion([("selectkbest_pca", make_pipeline(SelectKBest(k=3), FeatureUnion([('pca', PCA(n_components=1)), ('nested_pipeline', make_pipeline(SelectKBest(k=2), Nystroem()))]))), ("nys", Nystroem(n_components=2, random_state=42))])        
        sklearn_pipeline = make_pipeline(union, KNeighborsClassifier())
        lale_pipeline = import_from_sklearn_pipeline(sklearn_pipeline)
        print(lale_pipeline.to_json())
        self.assertEqual(len(lale_pipeline.edges()), 8)
        #These assertions assume topological sort, which may not be unique. So the assertions are brittle.
        from lale.lib.sklearn.pca import PCAImpl
        from lale.lib.sklearn.nystroem import NystroemImpl
        from lale.lib.lale.concat_features import ConcatFeaturesImpl
        from lale.lib.sklearn.k_neighbors_classifier import KNeighborsClassifierImpl
        self.assertIsInstance(lale_pipeline.edges()[0][0]._impl, SelectKBest)
        self.assertIsInstance(lale_pipeline.edges()[0][1]._impl, PCAImpl)
        self.assertIsInstance(lale_pipeline.edges()[1][0]._impl, SelectKBest)
        self.assertIsInstance(lale_pipeline.edges()[1][1]._impl, SelectKBest)
        self.assertIsInstance(lale_pipeline.edges()[2][0]._impl, SelectKBest)
        self.assertIsInstance(lale_pipeline.edges()[2][1]._impl, NystroemImpl)
        self.assertIsInstance(lale_pipeline.edges()[3][0]._impl, PCAImpl)
        self.assertIsInstance(lale_pipeline.edges()[3][1]._impl, ConcatFeaturesImpl)
        self.assertIsInstance(lale_pipeline.edges()[4][0]._impl, NystroemImpl)
        self.assertIsInstance(lale_pipeline.edges()[4][1]._impl, ConcatFeaturesImpl)
        self.assertIsInstance(lale_pipeline.edges()[5][0]._impl, ConcatFeaturesImpl)
        self.assertIsInstance(lale_pipeline.edges()[5][1]._impl, ConcatFeaturesImpl)
        self.assertIsInstance(lale_pipeline.edges()[6][0]._impl, NystroemImpl)
        self.assertIsInstance(lale_pipeline.edges()[6][1]._impl, ConcatFeaturesImpl)
        self.assertIsInstance(lale_pipeline.edges()[7][0]._impl, ConcatFeaturesImpl)
        self.assertIsInstance(lale_pipeline.edges()[7][1]._impl, KNeighborsClassifierImpl)
        self.assert_equal_predictions(sklearn_pipeline, lale_pipeline)

    def test_import_from_sklearn_pipeline_nested_pipeline2(self):
        from sklearn.pipeline import FeatureUnion, make_pipeline       
        from sklearn.decomposition import PCA
        from sklearn.kernel_approximation import Nystroem
        from sklearn.feature_selection import SelectKBest
        from sklearn.neighbors import KNeighborsClassifier
        from sklearn.pipeline import make_pipeline
        union = FeatureUnion([("selectkbest_pca", make_pipeline(SelectKBest(k=3), make_pipeline(SelectKBest(k=2), PCA()))), ("nys", Nystroem(n_components=2, random_state=42))])        
        sklearn_pipeline = make_pipeline(union, KNeighborsClassifier())
        lale_pipeline = import_from_sklearn_pipeline(sklearn_pipeline)
        self.assertEqual(len(lale_pipeline.edges()), 5)
        from lale.lib.sklearn.pca import PCAImpl
        from lale.lib.sklearn.nystroem import NystroemImpl
        from lale.lib.lale.concat_features import ConcatFeaturesImpl
        from lale.lib.sklearn.k_neighbors_classifier import KNeighborsClassifierImpl
        self.assertIsInstance(lale_pipeline.edges()[0][0]._impl, SelectKBest)
        self.assertIsInstance(lale_pipeline.edges()[0][1]._impl, SelectKBest)
        self.assertIsInstance(lale_pipeline.edges()[1][0]._impl, SelectKBest)
        self.assertIsInstance(lale_pipeline.edges()[1][1]._impl, PCAImpl)
        self.assertIsInstance(lale_pipeline.edges()[2][0]._impl, PCAImpl)
        self.assertIsInstance(lale_pipeline.edges()[2][1]._impl, ConcatFeaturesImpl)
        self.assertIsInstance(lale_pipeline.edges()[3][0]._impl, NystroemImpl)
        self.assertIsInstance(lale_pipeline.edges()[3][1]._impl, ConcatFeaturesImpl)
        self.assertIsInstance(lale_pipeline.edges()[4][0]._impl, ConcatFeaturesImpl)
        self.assertIsInstance(lale_pipeline.edges()[4][1]._impl, KNeighborsClassifierImpl)

        self.assert_equal_predictions(sklearn_pipeline, lale_pipeline)

    def test_export_to_sklearn_pipeline(self):
        from lale.lib.sklearn import PCA
        from lale.lib.sklearn import KNeighborsClassifier
        from sklearn.pipeline import make_pipeline
        lale_pipeline = PCA(n_components=3) >>  KNeighborsClassifier()
        trained_lale_pipeline = lale_pipeline.fit(self.X_train, self.y_train)
        sklearn_pipeline = trained_lale_pipeline.export_to_sklearn_pipeline()
        for i, pipeline_step in enumerate(sklearn_pipeline.named_steps):
            sklearn_step_params = sklearn_pipeline.named_steps[pipeline_step].get_params()
            lale_sklearn_params = trained_lale_pipeline.steps()[i]._impl._sklearn_model.get_params()
            self.assertEqual(sklearn_step_params, lale_sklearn_params)
        self.assert_equal_predictions(sklearn_pipeline, trained_lale_pipeline)

    def test_export_to_sklearn_pipeline1(self):
        from lale.lib.sklearn import PCA
        from lale.lib.sklearn import KNeighborsClassifier
        from sklearn.feature_selection import SelectKBest
        from sklearn.pipeline import make_pipeline
        lale_pipeline = SelectKBest(k=3) >>  KNeighborsClassifier()
        trained_lale_pipeline = lale_pipeline.fit(self.X_train, self.y_train)
        sklearn_pipeline = trained_lale_pipeline.export_to_sklearn_pipeline()
        for i, pipeline_step in enumerate(sklearn_pipeline.named_steps):
            sklearn_step_params = type(sklearn_pipeline.named_steps[pipeline_step])
            lale_sklearn_params = type(trained_lale_pipeline.steps()[i]._impl._sklearn_model) if hasattr(trained_lale_pipeline.steps()[i]._impl, '_sklearn_model') else type(trained_lale_pipeline.steps()[i]._impl)                
            self.assertEqual(sklearn_step_params, lale_sklearn_params)
        self.assert_equal_predictions(sklearn_pipeline, trained_lale_pipeline)

    def test_export_to_sklearn_pipeline2(self):
        from lale.lib.lale import ConcatFeatures
        from lale.lib.sklearn import PCA
        from lale.lib.sklearn import KNeighborsClassifier
        from sklearn.feature_selection import SelectKBest
        from lale.lib.sklearn import Nystroem
        from sklearn.pipeline import FeatureUnion

        lale_pipeline = (((PCA() & SelectKBest(k=3)) >> ConcatFeatures()) & Nystroem()) >> ConcatFeatures() >> KNeighborsClassifier()
        trained_lale_pipeline = lale_pipeline.fit(self.X_train, self.y_train)
        sklearn_pipeline = trained_lale_pipeline.export_to_sklearn_pipeline()
        self.assertIsInstance(sklearn_pipeline.named_steps['featureunion'], FeatureUnion)
        from sklearn.neighbors import KNeighborsClassifier
        self.assertIsInstance(sklearn_pipeline.named_steps['kneighborsclassifier'], KNeighborsClassifier)
        self.assert_equal_predictions(sklearn_pipeline, trained_lale_pipeline)

    def test_export_to_sklearn_pipeline3(self):
        from lale.lib.lale import ConcatFeatures
        from lale.lib.sklearn import PCA
        from lale.lib.sklearn import KNeighborsClassifier, LogisticRegression, SVC 
        from sklearn.feature_selection import SelectKBest
        from lale.lib.sklearn import Nystroem
        from sklearn.pipeline import FeatureUnion

        lale_pipeline = ((PCA() >> SelectKBest(k=2)) & (Nystroem(random_state = 42) >> SelectKBest(k=3))
         & (SelectKBest(k=3))) >> ConcatFeatures() >> SelectKBest(k=2) >> LogisticRegression()
        trained_lale_pipeline = lale_pipeline.fit(self.X_train, self.y_train)
        sklearn_pipeline = trained_lale_pipeline.export_to_sklearn_pipeline()
        self.assertIsInstance(sklearn_pipeline.named_steps['featureunion'], FeatureUnion)
        self.assertIsInstance(sklearn_pipeline.named_steps['selectkbest'], SelectKBest)
        from sklearn.linear_model import LogisticRegression
        self.assertIsInstance(sklearn_pipeline.named_steps['logisticregression'], LogisticRegression)
        self.assert_equal_predictions(sklearn_pipeline, trained_lale_pipeline)

    def test_export_to_sklearn_pipeline4(self):
        from lale.lib.sklearn import LogisticRegression
        from lale.operators import make_pipeline

        lale_pipeline = make_pipeline(LogisticRegression())
        trained_lale_pipeline = lale_pipeline.fit(self.X_train, self.y_train)
        sklearn_pipeline = trained_lale_pipeline.export_to_sklearn_pipeline()
        from sklearn.linear_model import LogisticRegression
        self.assertIsInstance(sklearn_pipeline.named_steps['logisticregression'], LogisticRegression)
        self.assert_equal_predictions(sklearn_pipeline, trained_lale_pipeline)

    #Doesn't work yet.
    # def test_export_to_pickle(self):
    #     from lale.lib.sklearn import LogisticRegression
    #     from lale.operators import make_pipeline

    #     lale_pipeline = make_pipeline(LogisticRegression())
    #     trained_lale_pipeline = lale_pipeline.fit(self.X_train, self.y_train)
    #     pickle.dumps(lale_pipeline)
    #     pickle.dumps(trained_lale_pipeline)

class TestComposition(unittest.TestCase):
    def setUp(self):
        from sklearn.datasets import load_iris
        from sklearn.model_selection import train_test_split
        data = load_iris()
        X, y = data.data, data.target
        self.X_train, self.X_test, self.y_train, self.y_test =  train_test_split(X, y)    
    def test_two_estimators_predict(self):
        pipeline = StandardScaler()  >> ( PCA() & Nystroem() & LogisticRegression() )>>ConcatFeatures() >> NoOp() >> LogisticRegression()
        trained = pipeline.fit(self.X_train, self.y_train)
        trained.predict(self.X_test)
    def test_two_estimators_predict1(self):
        pipeline = StandardScaler()  >> ( PCA() & Nystroem() & PassiveAggressiveClassifier() )>>ConcatFeatures() >> NoOp() >> PassiveAggressiveClassifier()
        trained = pipeline.fit(self.X_train, self.y_train)
        trained.predict(self.X_test)
    def test_two_estimators_predict_proba(self):
        pipeline = StandardScaler()  >> ( PCA() & Nystroem() & LogisticRegression() )>>ConcatFeatures() >> NoOp() >> LogisticRegression()
        pipeline.fit(self.X_train, self.y_train)
        pipeline.predict_proba(self.X_test)
    def test_two_estimators_predict_proba1(self):
        pipeline = StandardScaler()  >> ( PCA() & Nystroem() & PassiveAggressiveClassifier() )>>ConcatFeatures() >> NoOp() >> PassiveAggressiveClassifier()
        pipeline.fit(self.X_train, self.y_train)
        with self.assertRaises(ValueError):
            pipeline.predict_proba(self.X_test)
    def test_multiple_estimators_predict_predict_proba(self) :
        pipeline = (
            StandardScaler() >>
            ( LogisticRegression() & PCA() ) >> ConcatFeatures() >>
            ( NoOp() & LinearSVC() ) >> ConcatFeatures() >>
            KNeighborsClassifier()
        )
        pipeline.fit(self.X_train, self.y_train)
        tmp = pipeline.predict_proba(self.X_test)
        tmp = pipeline.predict(self.X_test)





from sklearn.preprocessing._function_transformer import FunctionTransformer as SKLModel
import lale.helpers
import lale.operators
from numpy import nan, inf

class FunctionTransformerImpl():

    def __init__(self, func=None, inverse_func=None, validate=None, accept_sparse=False, pass_y='deprecated', check_inverse=True, kw_args=None, inv_kw_args=None):
        self._hyperparams = {
            'func': func,
            'inverse_func': inverse_func,
            'validate': validate,
            'accept_sparse': accept_sparse,
            'pass_y': pass_y,
            'check_inverse': check_inverse,
            'kw_args': kw_args,
            'inv_kw_args': inv_kw_args}

    def fit(self, X, y=None):
        self._sklearn_model = SKLModel(**self._hyperparams)
        if (y is not None):
            self._sklearn_model.fit(X, y)
        else:
            self._sklearn_model.fit(X)
        return self

    def transform(self, X):
        return self._sklearn_model.transform(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for FunctionTransformer    Constructs a transformer from an arbitrary callable.',
    'allOf': [{
        'type': 'object',
        'required': ['func', 'inverse_func', 'validate', 'accept_sparse', 'pass_y', 'check_inverse', 'kw_args', 'inv_kw_args'],
        'relevantToOptimizer': ['accept_sparse', 'pass_y'],
        'additionalProperties': False,
        'properties': {
            'func': {
                'anyOf': [{
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The callable to use for the transformation. This will be passed'},
            'inverse_func': {
                'anyOf': [{
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'The callable to use for the inverse transformation. This will be'},
            'validate': {
                'anyOf': [{
                    'type': 'boolean'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Indicate that the input X array should be checked before calling'},
            'accept_sparse': {
                'type': 'boolean',
                'default': False,
                'description': 'Indicate that func accepts a sparse matrix as input. If validate is'},
            'pass_y': {
                'anyOf': [{
                    'type': 'boolean'}, {
                    'enum': ['deprecated']}],
                'default': 'deprecated',
                'description': 'Indicate that transform should forward the y argument to the'},
            'check_inverse': {
                'type': 'boolean',
                'default': True,
                'description': 'Whether to check that or ``func`` followed by ``inverse_func`` leads to'},
            'kw_args': {
                'anyOf': [{
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Dictionary of additional keyword arguments to pass to func.'},
            'inv_kw_args': {
                'anyOf': [{
                    'type': 'object'}, {
                    'enum': [None]}],
                'default': None,
                'description': 'Dictionary of additional keyword arguments to pass to inverse_func.'},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit transformer by checking X.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Input array.'},
    },
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Transform X using the forward function.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'Input array.'},
        'y': {
            'XXX TODO XXX': '(ignored)',
            'description': '.. deprecated::0.19'},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Transformed input.',
    'type': 'array',
    'items': {
        'type': 'array',
        'items': {
            'type': 'number'},
    },
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema},
}
if (__name__ == '__main__'):
    lale.helpers.validate_is_schema(_combined_schemas)
FunctionTransformer = lale.operators.make_operator(FunctionTransformerImpl, _combined_schemas)





from sklearn.preprocessing._encoders import OrdinalEncoder as SKLModel
import lale.helpers
import lale.operators
from numpy import nan, inf

class OrdinalEncoderImpl():

    def __init__(self, categories='auto', dtype=None):
        self._hyperparams = {
            'categories': categories,
            'dtype': dtype}

    def fit(self, X, y=None):
        self._sklearn_model = SKLModel(**self._hyperparams)
        if (y is not None):
            self._sklearn_model.fit(X, y)
        else:
            self._sklearn_model.fit(X)
        return self

    def transform(self, X):
        return self._sklearn_model.transform(X)
_hyperparams_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'inherited docstring for OrdinalEncoder    Encode categorical features as an integer array.',
    'allOf': [{
        'type': 'object',
        'required': ['categories', 'dtype'],
        'relevantToOptimizer': [],
        'additionalProperties': False,
        'properties': {
            'categories': {
                'XXX TODO XXX': "'auto' or a list of lists/arrays of values.",
                'description': 'Categories (unique values) per feature:',
                'enum': ['auto'],
                'default': 'auto'},
            'dtype': {
                'XXX TODO XXX': 'number type, default np.float64',
                'description': 'Desired dtype of output.'},
        }}],
}
_input_fit_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Fit the OrdinalEncoder to X.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'The data to determine the categories of each feature.'},
    },
}
_input_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Transform X to ordinal codes.',
    'type': 'object',
    'properties': {
        'X': {
            'type': 'array',
            'items': {
                'type': 'array',
                'items': {
                    'type': 'number'},
            },
            'description': 'The data to encode.'},
    },
}
_output_transform_schema = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Transformed input.',
    'XXX TODO XXX': 'sparse matrix or a 2-d array',
}
_combined_schemas = {
    '$schema': 'http://json-schema.org/draft-04/schema#',
    'description': 'Combined schema for expected data and hyperparameters.',
    'type': 'object',
    'tags': {
        'pre': [],
        'op': ['transformer'],
        'post': []},
    'properties': {
        'hyperparams': _hyperparams_schema,
        'input_fit': _input_fit_schema,
        'input_transform': _input_transform_schema,
        'output_transform': _output_transform_schema},
}
if (__name__ == '__main__'):
    lale.helpers.validate_is_schema(_combined_schemas)
OrdinalEncoder = lale.operators.make_operator(OrdinalEncoderImpl, _combined_schemas)

